Struct Duration
pub struct Duration { /* private fields */ }Expand description
Defines generally usable durations for nanosecond precision valid for 32,768 centuries in either direction, and only on 80 bits / 10 octets.
Important conventions:
- The negative durations can be mentally modeled “BC” years. One hours before 01 Jan 0000, it was “-1” years but 365 days and 23h into the current day. It was decided that the nanoseconds corresponds to the nanoseconds into the current century. In other words, a duration with centuries = -1 and nanoseconds = 0 is a greater duration (further from zero) than centuries = -1 and nanoseconds = 1. Duration zero minus one nanosecond returns a century of -1 and a nanosecond set to the number of nanoseconds in one century minus one. That difference is exactly 1 nanoseconds, where the former duration is “closer to zero” than the latter. As such, the largest negative duration that can be represented sets the centuries to i16::MAX and its nanoseconds to NANOSECONDS_PER_CENTURY.
- Negative and positive durations are distinct: -15 minutes != 15 minutes. Use the signum function to check the sign, and abs() to get the absolute value.
:type string_repr: str
Implementations
impl Duration
impl Duration
pub const fn py_to_parts(&self) -> (i16, u64)
pub const fn py_to_parts(&self) -> (i16, u64)
Returns the centuries and nanoseconds of this duration NOTE: These items are not public to prevent incorrect durations from being created by modifying the values of the structure directly. :rtype: tuple[int, int]
pub fn py_total_nanoseconds(&self) -> i128
pub fn py_total_nanoseconds(&self) -> i128
Returns the total nanoseconds in a signed 128 bit integer :rtype: int
pub fn py_to_seconds(&self) -> f64
pub fn py_to_seconds(&self) -> f64
Returns this duration in seconds f64. For high fidelity comparisons, it is recommended to keep using the Duration structure. :rtype: float
pub fn py_to_unit(&self, unit: Unit) -> f64
pub fn py_to_unit(&self, unit: Unit) -> f64
:type unit: Unit :rtype: float
pub const fn py_signum(&self) -> i8
pub const fn py_signum(&self) -> i8
Returns the sign of this duration
- 0 if the number is zero
- 1 if the number is positive
- -1 if the number is negative :rtype: int
pub fn py_decompose(&self) -> (i8, u64, u64, u64, u64, u64, u64, u64)
pub fn py_decompose(&self) -> (i8, u64, u64, u64, u64, u64, u64, u64)
Decomposes a Duration in its sign, days, hours, minutes, seconds, ms, us, ns
:rtype: tuple[int, int, int, int, int, int, int, int]
pub fn py_floor(&self, duration: Duration) -> Duration
pub fn py_floor(&self, duration: Duration) -> Duration
Floors this duration to the closest duration from the bottom
Example
use hifitime::{Duration, TimeUnits};
let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.floor(1.hours()), 2.hours());
assert_eq!(two_hours_three_min.floor(30.minutes()), 2.hours());
// This is zero because we floor by a duration longer than the current duration, rounding it down
assert_eq!(two_hours_three_min.floor(4.hours()), 0.hours());
assert_eq!(two_hours_three_min.floor(1.seconds()), two_hours_three_min);
assert_eq!(two_hours_three_min.floor(1.hours() + 1.minutes()), 2.hours() + 2.minutes());
assert_eq!(two_hours_three_min.floor(1.hours() + 5.minutes()), 1.hours() + 5.minutes());:type duration: Duration :rtype: Duration
pub fn py_ceil(&self, duration: Duration) -> Duration
pub fn py_ceil(&self, duration: Duration) -> Duration
Ceils this duration to the closest provided duration
This simply floors then adds the requested duration
Example
use hifitime::{Duration, TimeUnits};
let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.ceil(1.hours()), 3.hours());
assert_eq!(two_hours_three_min.ceil(30.minutes()), 2.hours() + 30.minutes());
assert_eq!(two_hours_three_min.ceil(4.hours()), 4.hours());
assert_eq!(two_hours_three_min.ceil(1.seconds()), two_hours_three_min + 1.seconds());
assert_eq!(two_hours_three_min.ceil(1.hours() + 5.minutes()), 2.hours() + 10.minutes());:type duration: Duration :rtype: Duration
pub fn py_round(&self, duration: Duration) -> Duration
pub fn py_round(&self, duration: Duration) -> Duration
Rounds this duration to the closest provided duration
This performs both a ceil and floor and returns the value which is the closest to current one.
Example
use hifitime::{Duration, TimeUnits};
let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.round(1.hours()), 2.hours());
assert_eq!(two_hours_three_min.round(30.minutes()), 2.hours());
assert_eq!(two_hours_three_min.round(4.hours()), 4.hours());
assert_eq!(two_hours_three_min.round(1.seconds()), two_hours_three_min);
assert_eq!(two_hours_three_min.round(1.hours() + 5.minutes()), 2.hours() + 10.minutes());:type duration: Duration :rtype: Duration
pub fn py_approx(&self) -> Duration
pub fn py_approx(&self) -> Duration
Rounds this duration to the largest units represented in this duration.
This is useful to provide an approximate human duration. Under the hood, this function uses round,
so the “tipping point” of the rounding is half way to the next increment of the greatest unit.
As shown below, one example is that 35 hours and 59 minutes rounds to 1 day, but 36 hours and 1 minute rounds
to 2 days because 2 days is closer to 36h 1 min than 36h 1 min is to 1 day.
Example
use hifitime::{Duration, TimeUnits};
assert_eq!((2.hours() + 3.minutes()).approx(), 2.hours());
assert_eq!((24.hours() + 3.minutes()).approx(), 1.days());
assert_eq!((35.hours() + 59.minutes()).approx(), 1.days());
assert_eq!((36.hours() + 1.minutes()).approx(), 2.days());
assert_eq!((47.hours() + 3.minutes()).approx(), 2.days());
assert_eq!((49.hours() + 3.minutes()).approx(), 2.days());:rtype: Duration
pub fn py_min(&self, other: Duration) -> Duration
pub fn py_min(&self, other: Duration) -> Duration
Returns the minimum of the two durations.
use hifitime::TimeUnits;
let d0 = 20.seconds();
let d1 = 21.seconds();
assert_eq!(d0, d1.min(d0));
assert_eq!(d0, d0.min(d1));:type other: Duration :rtype: Duration
pub fn py_max(&self, other: Duration) -> Duration
pub fn py_max(&self, other: Duration) -> Duration
Returns the maximum of the two durations.
use hifitime::TimeUnits;
let d0 = 20.seconds();
let d1 = 21.seconds();
assert_eq!(d1, d1.max(d0));
assert_eq!(d1, d0.max(d1));:type other: Duration :rtype: Duration
pub fn py_is_negative(&self) -> bool
pub fn py_is_negative(&self) -> bool
Returns whether this is a negative or positive duration. :rtype: bool
impl Duration
impl Duration
pub const MIN_POSITIVE: Duration = Self::EPSILON
pub const MIN_POSITIVE: Duration = Self::EPSILON
Minimum positive duration is one nanoseconds
pub const MIN_NEGATIVE: Duration
pub const MIN_NEGATIVE: Duration
Minimum negative duration is minus one nanosecond
pub const fn from_parts(centuries: i16, nanoseconds: u64) -> Duration
pub const fn from_parts(centuries: i16, nanoseconds: u64) -> Duration
Create a normalized duration from its parts
pub const fn from_total_nanoseconds(nanos: i128) -> Duration
pub const fn from_total_nanoseconds(nanos: i128) -> Duration
Converts the total nanoseconds as i128 into this Duration (saving 48 bits)
pub const fn from_truncated_nanoseconds(nanos: i64) -> Duration
pub const fn from_truncated_nanoseconds(nanos: i64) -> Duration
Create a new duration from the truncated nanoseconds (+/- 2927.1 years of duration)
pub const fn from_days(value: f64) -> Duration
pub const fn from_days(value: f64) -> Duration
Creates a new duration from the provided number of days
pub const fn from_hours(value: f64) -> Duration
pub const fn from_hours(value: f64) -> Duration
Creates a new duration from the provided number of hours
pub const fn from_seconds(value: f64) -> Duration
pub const fn from_seconds(value: f64) -> Duration
Creates a new duration from the provided number of seconds
pub const fn from_milliseconds(value: f64) -> Duration
pub const fn from_milliseconds(value: f64) -> Duration
Creates a new duration from the provided number of milliseconds
pub const fn from_microseconds(value: f64) -> Duration
pub const fn from_microseconds(value: f64) -> Duration
Creates a new duration from the provided number of microsecond
pub const fn from_nanoseconds(value: f64) -> Duration
pub const fn from_nanoseconds(value: f64) -> Duration
Creates a new duration from the provided number of nanoseconds
pub fn compose(
sign: i8,
days: u64,
hours: u64,
minutes: u64,
seconds: u64,
milliseconds: u64,
microseconds: u64,
nanoseconds: u64,
) -> Duration
pub fn compose( sign: i8, days: u64, hours: u64, minutes: u64, seconds: u64, milliseconds: u64, microseconds: u64, nanoseconds: u64, ) -> Duration
Creates a new duration from its parts. Set the sign to a negative number for the duration to be negative.
pub fn compose_f64(
sign: i8,
days: f64,
hours: f64,
minutes: f64,
seconds: f64,
milliseconds: f64,
microseconds: f64,
nanoseconds: f64,
) -> Duration
pub fn compose_f64( sign: i8, days: f64, hours: f64, minutes: f64, seconds: f64, milliseconds: f64, microseconds: f64, nanoseconds: f64, ) -> Duration
Creates a new duration from its parts. Set the sign to a negative number for the duration to be negative.
pub fn from_tz_offset(sign: i8, hours: i64, minutes: i64) -> Duration
pub fn from_tz_offset(sign: i8, hours: i64, minutes: i64) -> Duration
Initializes a Duration from a timezone offset
impl Duration
impl Duration
pub const fn to_parts(&self) -> (i16, u64)
pub const fn to_parts(&self) -> (i16, u64)
Returns the centuries and nanoseconds of this duration NOTE: These items are not public to prevent incorrect durations from being created by modifying the values of the structure directly.
pub fn total_nanoseconds(&self) -> i128
pub fn total_nanoseconds(&self) -> i128
Returns the total nanoseconds in a signed 128 bit integer
pub fn try_truncated_nanoseconds(&self) -> Result<i64, HifitimeError>
pub fn try_truncated_nanoseconds(&self) -> Result<i64, HifitimeError>
Returns the truncated nanoseconds in a signed 64 bit integer, if the duration fits.
pub fn truncated_nanoseconds(&self) -> i64
pub fn truncated_nanoseconds(&self) -> i64
Returns the truncated nanoseconds in a signed 64 bit integer, if the duration fits. WARNING: This function will NOT fail and will return the i64::MIN or i64::MAX depending on the sign of the centuries if the Duration does not fit on aa i64
pub fn to_seconds(&self) -> f64
pub fn to_seconds(&self) -> f64
Returns this duration in seconds f64. For high fidelity comparisons, it is recommended to keep using the Duration structure.
Examples found in repository?
26fn main() -> Result<(), Box<dyn Error>> {
27 pel::init();
28 // Dynamics models require planetary constants and ephemerides to be defined.
29 // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
30 // This will automatically download the DE440s planetary ephemeris,
31 // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
32 // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
33 // planetary constants kernels.
34 // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
35 // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
36 // references to many functions.
37 let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
38 // Define the orbit epoch
39 let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
40
41 // Define the orbit.
42 // First we need to fetch the Earth J2000 from information from the Almanac.
43 // This allows the frame to include the gravitational parameters and the shape of the Earth,
44 // defined as a tri-axial ellipoid. Note that this shape can be changed manually or in the Almanac
45 // by loading a different set of planetary constants.
46 let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
47
48 // Placing this GEO bird just above Colorado.
49 // In theory, the eccentricity is zero, but in practice, it's about 1e-5 to 1e-6 at best.
50 let orbit = Orbit::try_keplerian(42164.0, 1e-5, 0., 163.0, 75.0, 0.0, epoch, earth_j2000)?;
51 // Print in in Keplerian form.
52 println!("{orbit:x}");
53
54 let state_bf = almanac.transform_to(orbit, IAU_EARTH_FRAME, None)?;
55 let (orig_lat_deg, orig_long_deg, orig_alt_km) = state_bf.latlongalt()?;
56
57 // Nyx is used for high fidelity propagation, not Keplerian propagation as above.
58 // Nyx only propagates Spacecraft at the moment, which allows it to account for acceleration
59 // models such as solar radiation pressure.
60
61 // Let's build a cubesat sized spacecraft, with an SRP area of 10 cm^2 and a mass of 9.6 kg.
62 let sc = Spacecraft::builder()
63 .orbit(orbit)
64 .mass(Mass::from_dry_mass(9.60))
65 .srp(SRPData {
66 area_m2: 10e-4,
67 coeff_reflectivity: 1.1,
68 })
69 .build();
70 println!("{sc:x}");
71
72 // Set up the spacecraft dynamics.
73
74 // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
75 // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
76 let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
77
78 // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
79 // We're using the JGM3 model here, which is the default in GMAT.
80 let mut jgm3_meta = MetaFile {
81 uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
82 crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
83 };
84 // And let's download it if we don't have it yet.
85 jgm3_meta.process(true)?;
86
87 // Build the spherical harmonics.
88 // The harmonics must be computed in the body fixed frame.
89 // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
90 let harmonics_21x21 = GravityField::new(
91 GravityFieldData::from_cof(&jgm3_meta.uri, 21, 21, almanac.frame_info(IAU_EARTH_FRAME)?)
92 .unwrap(),
93 );
94
95 // Include the spherical harmonics into the orbital dynamics.
96 orbital_dyn.accel_models.push(harmonics_21x21);
97
98 // We define the solar radiation pressure, using the default solar flux and accounting only
99 // for the eclipsing caused by the Earth and Moon.
100 let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
101
102 // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
103 // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
104 let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
105
106 println!("{dynamics}");
107
108 // Finally, let's propagate this orbit to the same epoch as above.
109 // The first returned value is the spacecraft state at the final epoch.
110 // The second value is the full trajectory where the step size is variable step used by the propagator.
111 let (future_sc, trajectory) = Propagator::default(dynamics)
112 .with(sc, almanac.clone())
113 .until_epoch_with_traj(epoch + Unit::Century * 0.03)?;
114
115 println!("=== High fidelity propagation ===");
116 println!(
117 "SMA changed by {:.3} km",
118 orbit.sma_km()? - future_sc.orbit.sma_km()?
119 );
120 println!(
121 "ECC changed by {:.6}",
122 orbit.ecc()? - future_sc.orbit.ecc()?
123 );
124 println!(
125 "INC changed by {:.3e} deg",
126 orbit.inc_deg()? - future_sc.orbit.inc_deg()?
127 );
128 println!(
129 "RAAN changed by {:.3} deg",
130 orbit.raan_deg()? - future_sc.orbit.raan_deg()?
131 );
132 println!(
133 "AOP changed by {:.3} deg",
134 orbit.aop_deg()? - future_sc.orbit.aop_deg()?
135 );
136 println!(
137 "TA changed by {:.3} deg",
138 orbit.ta_deg()? - future_sc.orbit.ta_deg()?
139 );
140
141 // We also have access to the full trajectory throughout the propagation.
142 println!("{trajectory}");
143
144 println!("Spacecraft params after 3 years without active control:\n{future_sc:x}");
145
146 // With the trajectory, let's build a few data products.
147
148 // 1. Export the trajectory as a parquet file, which includes the Keplerian orbital elements.
149
150 let analysis_step = Unit::Minute * 5;
151
152 trajectory.to_parquet(
153 "./03_geo_hf_prop.parquet",
154 ExportCfg::builder().step(analysis_step).build(),
155 )?;
156
157 // 2. Compute the latitude, longitude, and altitude throughout the trajectory by rotating the spacecraft position into the Earth body fixed frame.
158
159 // We iterate over the trajectory, grabbing a state every two minutes.
160 let mut offset_s = vec![];
161 let mut epoch_str = vec![];
162 let mut longitude_deg = vec![];
163 let mut latitude_deg = vec![];
164 let mut altitude_km = vec![];
165
166 for state in trajectory.every(analysis_step) {
167 // Convert the GEO bird state into the body fixed frame, and keep track of its latitude, longitude, and altitude.
168 // These define the GEO stationkeeping box.
169
170 let this_epoch = state.epoch();
171
172 offset_s.push((this_epoch - orbit.epoch).to_seconds());
173 epoch_str.push(this_epoch.to_isoformat());
174
175 let state_bf = almanac.transform_to(state.orbit, IAU_EARTH_FRAME, None)?;
176 let (lat_deg, long_deg, alt_km) = state_bf.latlongalt()?;
177 longitude_deg.push(long_deg);
178 latitude_deg.push(lat_deg);
179 altitude_km.push(alt_km);
180 }
181
182 println!(
183 "Longitude changed by {:.3} deg -- Box is 0.1 deg E-W",
184 orig_long_deg - longitude_deg.last().unwrap()
185 );
186
187 println!(
188 "Latitude changed by {:.3} deg -- Box is 0.05 deg N-S",
189 orig_lat_deg - latitude_deg.last().unwrap()
190 );
191
192 println!(
193 "Altitude changed by {:.3} km -- Box is 30 km",
194 orig_alt_km - altitude_km.last().unwrap()
195 );
196
197 // Build the station keeping data frame.
198 let mut sk_df = df!(
199 "Offset (s)" => offset_s.clone(),
200 "Epoch (UTC)" => epoch_str.clone(),
201 "Longitude E-W (deg)" => longitude_deg,
202 "Latitude N-S (deg)" => latitude_deg,
203 "Altitude (km)" => altitude_km,
204
205 )?;
206
207 // Create a file to write the Parquet to
208 let file = File::create("./03_geo_lla.parquet").expect("Could not create file");
209
210 // Create a ParquetWriter and write the DataFrame to the file
211 ParquetWriter::new(file).finish(&mut sk_df)?;
212
213 Ok(())
214}More examples
30fn main() -> Result<(), Box<dyn Error>> {
31 pel::init();
32 // Dynamics models require planetary constants and ephemerides to be defined.
33 // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
34 // This will automatically download the DE440s planetary ephemeris,
35 // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
36 // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
37 // planetary constants kernels.
38 // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
39 // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
40 // references to many functions.
41 let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
42 // Define the orbit epoch
43 let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
44
45 // Define the orbit.
46 // First we need to fetch the Earth J2000 from information from the Almanac.
47 // This allows the frame to include the gravitational parameters and the shape of the Earth,
48 // defined as a tri-axial ellipoid. Note that this shape can be changed manually or in the Almanac
49 // by loading a different set of planetary constants.
50 let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
51
52 let orbit =
53 Orbit::try_keplerian_altitude(300.0, 0.015, 68.5, 65.2, 75.0, 0.0, epoch, earth_j2000)?;
54 // Print in in Keplerian form.
55 println!("{orbit:x}");
56
57 // There are two ways to propagate an orbit. We can make a quick approximation assuming only two-body
58 // motion. This is a useful first order approximation but it isn't used in real-world applications.
59
60 // This approach is a feature of ANISE.
61 let future_orbit_tb = orbit.at_epoch(epoch + Unit::Day * 3)?;
62 println!("{future_orbit_tb:x}");
63
64 // Two body propagation relies solely on Kepler's laws, so only the true anomaly will change.
65 println!(
66 "SMA changed by {:.3e} km",
67 orbit.sma_km()? - future_orbit_tb.sma_km()?
68 );
69 println!(
70 "ECC changed by {:.3e}",
71 orbit.ecc()? - future_orbit_tb.ecc()?
72 );
73 println!(
74 "INC changed by {:.3e} deg",
75 orbit.inc_deg()? - future_orbit_tb.inc_deg()?
76 );
77 println!(
78 "RAAN changed by {:.3e} deg",
79 orbit.raan_deg()? - future_orbit_tb.raan_deg()?
80 );
81 println!(
82 "AOP changed by {:.3e} deg",
83 orbit.aop_deg()? - future_orbit_tb.aop_deg()?
84 );
85 println!(
86 "TA changed by {:.3} deg",
87 orbit.ta_deg()? - future_orbit_tb.ta_deg()?
88 );
89
90 // Nyx is used for high fidelity propagation, not Keplerian propagation as above.
91 // Nyx only propagates Spacecraft at the moment, which allows it to account for acceleration
92 // models such as solar radiation pressure.
93
94 // Let's build a cubesat sized spacecraft, with an SRP area of 10 cm^2 and a mass of 9.6 kg.
95 let sc = Spacecraft::builder()
96 .orbit(orbit)
97 .mass(Mass::from_dry_mass(9.60))
98 .srp(SRPData {
99 area_m2: 10e-4,
100 coeff_reflectivity: 1.1,
101 })
102 .build();
103 println!("{sc:x}");
104
105 // Set up the spacecraft dynamics.
106
107 // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
108 // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
109 let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
110
111 // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
112 // We're using the JGM3 model here, which is the default in GMAT.
113 let mut jgm3_meta = MetaFile {
114 uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
115 crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
116 };
117 // And let's download it if we don't have it yet.
118 jgm3_meta.process(true)?;
119
120 // Build the spherical harmonics.
121 // The harmonics must be computed in the body fixed frame.
122 // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
123 let harmonics_21x21 = GravityField::new(
124 GravityFieldData::from_cof(&jgm3_meta.uri, 21, 21, almanac.frame_info(IAU_EARTH_FRAME)?)
125 .unwrap(),
126 );
127
128 // Include the spherical harmonics into the orbital dynamics.
129 orbital_dyn.accel_models.push(harmonics_21x21);
130
131 // We define the solar radiation pressure, using the default solar flux and accounting only
132 // for the eclipsing caused by the Earth.
133 let srp_dyn = SolarPressure::default_flux(EARTH_J2000, &almanac)?;
134
135 // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
136 // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
137 let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
138
139 println!("{dynamics}");
140
141 // Finally, let's propagate this orbit to the same epoch as above.
142 // The first returned value is the spacecraft state at the final epoch.
143 // The second value is the full trajectory where the step size is variable step used by the propagator.
144 let (future_sc, trajectory) = Propagator::default(dynamics)
145 .with(sc, almanac.clone())
146 .until_epoch_with_traj(future_orbit_tb.epoch)?;
147
148 println!("=== High fidelity propagation ===");
149 println!(
150 "SMA changed by {:.3} km",
151 orbit.sma_km()? - future_sc.orbit.sma_km()?
152 );
153 println!(
154 "ECC changed by {:.6}",
155 orbit.ecc()? - future_sc.orbit.ecc()?
156 );
157 println!(
158 "INC changed by {:.3e} deg",
159 orbit.inc_deg()? - future_sc.orbit.inc_deg()?
160 );
161 println!(
162 "RAAN changed by {:.3} deg",
163 orbit.raan_deg()? - future_sc.orbit.raan_deg()?
164 );
165 println!(
166 "AOP changed by {:.3} deg",
167 orbit.aop_deg()? - future_sc.orbit.aop_deg()?
168 );
169 println!(
170 "TA changed by {:.3} deg",
171 orbit.ta_deg()? - future_sc.orbit.ta_deg()?
172 );
173
174 // We also have access to the full trajectory throughout the propagation.
175 println!("{trajectory}");
176
177 // With the trajectory, let's build a few data products.
178
179 // 1. Export the trajectory as a CCSDS OEM version 2.0 file and as a parquet file, which includes the Keplerian orbital elements.
180
181 trajectory.to_oem_file(
182 "./01_cubesat_hf_prop.oem",
183 "CUBESAT-ID".to_string(),
184 Some("Nyx Space".to_string()),
185 Some("CUBESAT".to_string()),
186 ExportCfg::builder().step(Unit::Minute * 2).build(),
187 )?;
188
189 trajectory.to_parquet_with_cfg(
190 "./01_cubesat_hf_prop.parquet",
191 ExportCfg::builder().step(Unit::Minute * 2).build(),
192 )?;
193
194 // 2. Compare the difference in the radial-intrack-crosstrack frame between the high fidelity
195 // and Keplerian propagation. The RIC frame is commonly used to compute the difference in position
196 // and velocity of different spacecraft.
197 // 3. Compute the azimuth, elevation, range, and range-rate data of that spacecraft as seen from Boulder, CO, USA.
198
199 let boulder_station = GroundStation::from_point(
200 "Boulder, CO, USA".to_string(),
201 40.014984, // latitude in degrees
202 -105.270546, // longitude in degrees
203 1.6550, // altitude in kilometers
204 almanac.frame_info(IAU_EARTH_FRAME)?,
205 );
206
207 // We iterate over the trajectory, grabbing a state every two minutes.
208 let mut offset_s = vec![];
209 let mut epoch_str = vec![];
210 let mut ric_x_km = vec![];
211 let mut ric_y_km = vec![];
212 let mut ric_z_km = vec![];
213 let mut ric_vx_km_s = vec![];
214 let mut ric_vy_km_s = vec![];
215 let mut ric_vz_km_s = vec![];
216
217 let mut azimuth_deg = vec![];
218 let mut elevation_deg = vec![];
219 let mut range_km = vec![];
220 let mut range_rate_km_s = vec![];
221 for state in trajectory.every(Unit::Minute * 2) {
222 // Try to compute the Keplerian/two body state just in time.
223 // This method occasionally fails to converge on an appropriate true anomaly
224 // from the mean anomaly. If that happens, we just skip this state.
225 // The high fidelity and Keplerian states diverge continuously, and we're curious
226 // about the divergence in this quick analysis.
227 let this_epoch = state.epoch();
228 match orbit.at_epoch(this_epoch) {
229 Ok(tb_then) => {
230 offset_s.push((this_epoch - orbit.epoch).to_seconds());
231 epoch_str.push(format!("{this_epoch}"));
232 // Compute the two body state just in time.
233 let ric = state.orbit.ric_difference(&tb_then)?;
234 ric_x_km.push(ric.radius_km.x);
235 ric_y_km.push(ric.radius_km.y);
236 ric_z_km.push(ric.radius_km.z);
237 ric_vx_km_s.push(ric.velocity_km_s.x);
238 ric_vy_km_s.push(ric.velocity_km_s.y);
239 ric_vz_km_s.push(ric.velocity_km_s.z);
240
241 // Compute the AER data for each state.
242 let aer = almanac.azimuth_elevation_range_sez(
243 state.orbit,
244 boulder_station.to_orbit(this_epoch, &almanac)?,
245 None,
246 None,
247 )?;
248 azimuth_deg.push(aer.azimuth_deg);
249 elevation_deg.push(aer.elevation_deg);
250 range_km.push(aer.range_km);
251 range_rate_km_s.push(aer.range_rate_km_s);
252 }
253 Err(e) => warn!("{} {e}", state.epoch()),
254 };
255 }
256
257 // Build the data frames.
258 let ric_df = df!(
259 "Offset (s)" => offset_s.clone(),
260 "Epoch" => epoch_str.clone(),
261 "RIC X (km)" => ric_x_km,
262 "RIC Y (km)" => ric_y_km,
263 "RIC Z (km)" => ric_z_km,
264 "RIC VX (km/s)" => ric_vx_km_s,
265 "RIC VY (km/s)" => ric_vy_km_s,
266 "RIC VZ (km/s)" => ric_vz_km_s,
267 )?;
268
269 println!("RIC difference at start\n{}", ric_df.head(Some(10)));
270 println!("RIC difference at end\n{}", ric_df.tail(Some(10)));
271
272 let aer_df = df!(
273 "Offset (s)" => offset_s.clone(),
274 "Epoch" => epoch_str.clone(),
275 "azimuth (deg)" => azimuth_deg,
276 "elevation (deg)" => elevation_deg,
277 "range (km)" => range_km,
278 "range rate (km/s)" => range_rate_km_s,
279 )?;
280
281 // Finally, let's see when the spacecraft is visible, assuming 15 degrees minimum elevation.
282 let mask = aer_df
283 .column("elevation (deg)")?
284 .gt(&Column::Scalar(ScalarColumn::new(
285 "elevation mask (deg)".into(),
286 Scalar::new(DataType::Float64, AnyValue::Float64(15.0)),
287 offset_s.len(),
288 )))?;
289 let cubesat_visible = aer_df.filter(&mask)?;
290
291 println!("{cubesat_visible}");
292
293 Ok(())
294}pub fn to_unit(&self, unit: Unit) -> f64
pub fn abs(&self) -> Duration
pub fn abs(&self) -> Duration
Returns the absolute value of this duration
Examples found in repository?
35fn main() -> Result<(), Box<dyn Error>> {
36 pel::init();
37
38 // ====================== //
39 // === ALMANAC SET UP === //
40 // ====================== //
41
42 // Dynamics models require planetary constants and ephemerides to be defined.
43 // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45 let output_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "../data", "04_output"]
46 .iter()
47 .collect();
48
49 let data_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "examples", "04_lro_od"]
50 .iter()
51 .collect();
52
53 let meta = data_folder.join("lro-dynamics.dhall");
54
55 // Load this ephem in the general Almanac we're using for this analysis.
56 let mut almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
57 .map_err(Box::new)?
58 .process(true)
59 .map_err(Box::new)?;
60
61 let mut moon_pc = almanac.get_planetary_data_from_id(MOON).unwrap();
62 moon_pc.mu_km3_s2 = 4902.74987;
63 almanac.set_planetary_data_from_id(MOON, moon_pc).unwrap();
64
65 let mut earth = almanac.get_planetary_data_from_id(EARTH).unwrap();
66 earth.mu_km3_s2 = 398600.436;
67 almanac.set_planetary_data_from_id(EARTH, earth).unwrap();
68
69 // Save this new kernel for reuse.
70 // In an operational context, this would be part of the "Lock" process, and should not change throughout the mission.
71 almanac
72 .planetary_data
73 .values()
74 .next()
75 .unwrap()
76 .save_as(&data_folder.join("lro-specific.pca"), true)?;
77
78 // Lock the almanac (an Arc is a read only structure).
79 let almanac = Arc::new(almanac);
80
81 // Orbit determination requires a Trajectory structure, which can be saved as parquet file.
82 // In our case, the trajectory comes from the BSP file, so we need to build a Trajectory from the almanac directly.
83 // To query the Almanac, we need to build the LRO frame in the J2000 orientation in our case.
84 // Inspecting the LRO BSP in the ANISE GUI shows us that NASA has assigned ID -85 to LRO.
85 let lro_frame = Frame::from_ephem_j2000(-85);
86
87 // To build the trajectory we need to provide a spacecraft template.
88 let sc_template = Spacecraft::builder()
89 .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0)) // Launch masses
90 .srp(SRPData {
91 // SRP configuration is arbitrary, but we will be estimating it anyway.
92 area_m2: 3.9 * 2.7,
93 coeff_reflectivity: 0.96,
94 })
95 .orbit(Orbit::zero(MOON_J2000)) // Setting a zero orbit here because it's just a template
96 .build();
97 // Now we can build the trajectory from the BSP file.
98 // We'll arbitrarily set the tracking arc to 24 hours with a five second time step.
99 let traj_as_flown = Traj::from_bsp(
100 lro_frame,
101 MOON_J2000,
102 &almanac,
103 sc_template,
104 5.seconds(),
105 Some(Epoch::from_str("2024-01-01 00:00:00 UTC")?),
106 Some(Epoch::from_str("2024-01-02 00:00:00 UTC")?),
107 Aberration::LT,
108 Some("LRO".to_string()),
109 )?;
110
111 println!("{traj_as_flown}");
112
113 // ====================== //
114 // === MODEL MATCHING === //
115 // ====================== //
116
117 // Set up the spacecraft dynamics.
118
119 // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
120 // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
121 let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
122
123 // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
124 // We're using the GRAIL JGGRX model.
125 let mut jggrx_meta = MetaFile {
126 uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
127 crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
128 };
129 // And let's download it if we don't have it yet.
130 jggrx_meta.process(true)?;
131
132 // Build the spherical harmonics.
133 // The harmonics must be computed in the body fixed frame.
134 // We're using the long term prediction of the Moon principal axes frame.
135 let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
136 let sph_harmonics = GravityField::new(GravityFieldData::from_shadr(
137 &jggrx_meta.uri,
138 80,
139 80,
140 almanac.frame_info(moon_pa_frame)?,
141 )?);
142
143 // Include the spherical harmonics into the orbital dynamics.
144 orbital_dyn.accel_models.push(sph_harmonics);
145
146 // We define the solar radiation pressure, using the default solar flux and accounting only
147 // for the eclipsing caused by the Earth and Moon.
148 // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
149 let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
150
151 // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
152 // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
153 let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
154
155 println!("{dynamics}");
156
157 // Now we can build the propagator.
158 let setup = Propagator::default_dp78(dynamics.clone());
159
160 // For reference, let's build the trajectory with Nyx's models from that LRO state.
161 let (sim_final, traj_as_sim) = setup
162 .with(*traj_as_flown.first(), almanac.clone())
163 .until_epoch_with_traj(traj_as_flown.last().epoch())?;
164
165 println!("SIM INIT: {:x}", traj_as_flown.first());
166 println!("SIM FINAL: {sim_final:x}");
167 // Compute RIC difference between SIM and LRO ephem
168 let sim_lro_delta = sim_final
169 .orbit
170 .ric_difference(&traj_as_flown.last().orbit)?;
171 println!("{traj_as_sim}");
172 println!(
173 "SIM v LRO - RIC Position (m): {:.3}",
174 sim_lro_delta.radius_km * 1e3
175 );
176 println!(
177 "SIM v LRO - RIC Velocity (m/s): {:.3}",
178 sim_lro_delta.velocity_km_s * 1e3
179 );
180
181 traj_as_sim.ric_diff_to_parquet(
182 &traj_as_flown,
183 output_folder.join("./04_lro_sim_truth_error.parquet"),
184 ExportCfg::default(),
185 )?;
186
187 // ==================== //
188 // === OD SIMULATOR === //
189 // ==================== //
190
191 // After quite some time trying to exactly match the model, we still end up with an oscillatory difference on the order of 150 meters between the propagated state
192 // and the truth LRO state.
193
194 // Therefore, we will actually run an estimation from a dispersed LRO state.
195 // The sc_seed is the true LRO state from the BSP.
196 let sc_seed = *traj_as_flown.first();
197
198 // Load the Deep Space Network ground stations.
199 // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
200 let ground_station_file: PathBuf = [
201 env!("CARGO_MANIFEST_DIR"),
202 "examples",
203 "04_lro_od",
204 "dsn-network.yaml",
205 ]
206 .iter()
207 .collect();
208
209 let devices = GroundStation::load_named(ground_station_file)?;
210
211 let mut proc_devices = devices.clone();
212
213 // Increase the noise in the devices to accept more measurements.
214 for gs in proc_devices.values_mut() {
215 if let Some(noise) = &mut gs
216 .stochastic_noises
217 .as_mut()
218 .unwrap()
219 .get_mut(&MeasurementType::Range)
220 {
221 *noise.white_noise.as_mut().unwrap() *= 3.0;
222 }
223 }
224
225 // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
226 // Nyx can build a tracking schedule for you based on the first station with access.
227 let trkconfg_yaml: PathBuf = [
228 env!("CARGO_MANIFEST_DIR"),
229 "examples",
230 "04_lro_od",
231 "tracking-cfg.yaml",
232 ]
233 .iter()
234 .collect();
235
236 let configs: BTreeMap<String, TrkConfig> = TrkConfig::load_named(trkconfg_yaml)?;
237
238 // Build the tracking arc simulation to generate a "standard measurement".
239 let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
240 devices.clone(),
241 traj_as_flown.clone(),
242 configs,
243 123, // Set a seed for reproducibility
244 )?;
245
246 trk.build_schedule(&almanac)?;
247 let arc = trk.generate_measurements(&almanac)?;
248 // Save the simulated tracking data
249 arc.to_parquet_simple(output_folder.join("04_lro_simulated_tracking.parquet"))?;
250
251 // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
252 println!("{arc}");
253
254 // Now that we have simulated measurements, we'll run the orbit determination.
255
256 // ===================== //
257 // === OD ESTIMATION === //
258 // ===================== //
259
260 let sc = SpacecraftUncertainty::builder()
261 .nominal(sc_seed)
262 .frame(LocalFrame::RIC)
263 .x_km(0.5)
264 .y_km(0.5)
265 .z_km(0.5)
266 .vx_km_s(5e-3)
267 .vy_km_s(5e-3)
268 .vz_km_s(5e-3)
269 .build();
270
271 // Build the filter initial estimate, which we will reuse in the filter.
272 let mut initial_estimate = sc.to_estimate()?;
273 initial_estimate.covar *= 3.0;
274
275 println!("== FILTER STATE ==\n{sc_seed:x}\n{initial_estimate}");
276
277 // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
278 let process_noise = ProcessNoise3D::from_velocity_km_s(
279 &[1e-12, 1e-12, 1e-12],
280 1 * Unit::Hour,
281 10 * Unit::Minute,
282 None,
283 );
284
285 println!("{process_noise}");
286
287 // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
288 let odp = SpacecraftKalmanOD::new(
289 setup,
290 KalmanVariant::ReferenceUpdate,
291 Some(SigmaRejection::default()),
292 proc_devices,
293 almanac.clone(),
294 )
295 .with_process_noise(process_noise);
296
297 let od_sol = odp.process_arc(initial_estimate, &arc)?;
298
299 let final_est = od_sol.estimates.last().unwrap();
300
301 println!("{final_est}");
302
303 let ric_err = traj_as_flown
304 .at(final_est.epoch())?
305 .orbit
306 .ric_difference(&final_est.orbital_state())?;
307 println!("== RIC at end ==");
308 println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
309 println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
310
311 println!(
312 "Num residuals rejected: #{}",
313 od_sol.rejected_residuals().len()
314 );
315 println!(
316 "Percentage within +/-3: {}",
317 od_sol.residual_ratio_within_threshold(3.0).unwrap()
318 );
319 println!("Ratios normal? {}", od_sol.is_normal(None).unwrap());
320
321 od_sol.to_parquet(
322 output_folder.join("04_lro_od_results.parquet"),
323 ExportCfg::default(),
324 )?;
325
326 // Create the ephemeris
327 let ephem = od_sol.to_ephemeris("LRO rebuilt".to_string());
328 let ephem_start = ephem.start_epoch().unwrap();
329 let ephem_end = ephem.end_epoch().unwrap();
330 // Check that the covariance is PSD throughout the ephemeris by interpolating it.
331 for epoch in TimeSeries::inclusive(ephem_start, ephem_end, Unit::Minute * 5) {
332 ephem
333 .covar_at(
334 epoch,
335 anise::ephemerides::ephemeris::LocalFrame::RIC,
336 &almanac,
337 )
338 .unwrap_or_else(|e| panic!("covar not PSD at {epoch}: {e}"));
339 }
340 // Export as BSP!
341 ephem
342 .write_spice_bsp(
343 -85,
344 output_folder.join("04_lro_rebuilt.bsp").to_str().unwrap(),
345 None,
346 )
347 .expect("could not built BSP");
348 let new_almanac = Almanac::default()
349 .load(output_folder.join("04_lro_rebuilt.bsp").to_str().unwrap())
350 .unwrap();
351 new_almanac.describe(None, None, None, None, None, None, None, None);
352 let (spk_start, spk_end) = new_almanac.spk_domain(-85).unwrap();
353
354 assert!((ephem_start - spk_start).abs() < Unit::Microsecond * 1);
355 assert!((ephem_end - spk_end).abs() < Unit::Microsecond * 1);
356
357 // In our case, we have the truth trajectory from NASA.
358 // So we can compute the RIC state difference between the real LRO ephem and what we've just estimated.
359 // Export the OD trajectory first.
360 let od_trajectory = od_sol.to_traj()?;
361 // Build the RIC difference.
362 od_trajectory.ric_diff_to_parquet(
363 &traj_as_flown,
364 output_folder.join("04_lro_od_truth_error.parquet"),
365 ExportCfg::default(),
366 )?;
367
368 Ok(())
369}pub const fn signum(&self) -> i8
pub const fn signum(&self) -> i8
Returns the sign of this duration
- 0 if the number is zero
- 1 if the number is positive
- -1 if the number is negative
pub fn decompose(&self) -> (i8, u64, u64, u64, u64, u64, u64, u64)
pub fn decompose(&self) -> (i8, u64, u64, u64, u64, u64, u64, u64)
Decomposes a Duration in its sign, days, hours, minutes, seconds, ms, us, ns
pub fn subdivision(&self, unit: Unit) -> Option<Duration>
pub fn subdivision(&self, unit: Unit) -> Option<Duration>
Returns the subdivision of duration in this unit, if such is available. Does not work with Week or Century.
Example
use hifitime::{Duration, TimeUnits, Unit};
let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.subdivision(Unit::Hour), Some(2.hours()));
assert_eq!(two_hours_three_min.subdivision(Unit::Minute), Some(3.minutes()));
assert_eq!(two_hours_three_min.subdivision(Unit::Second), Some(Duration::ZERO));
assert_eq!(two_hours_three_min.subdivision(Unit::Week), None);pub fn floor(&self, duration: Duration) -> Duration
pub fn floor(&self, duration: Duration) -> Duration
Floors this duration to the closest duration from the bottom
Example
use hifitime::{Duration, TimeUnits};
let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.floor(1.hours()), 2.hours());
assert_eq!(two_hours_three_min.floor(30.minutes()), 2.hours());
// This is zero because we floor by a duration longer than the current duration, rounding it down
assert_eq!(two_hours_three_min.floor(4.hours()), 0.hours());
assert_eq!(two_hours_three_min.floor(1.seconds()), two_hours_three_min);
assert_eq!(two_hours_three_min.floor(1.hours() + 1.minutes()), 2.hours() + 2.minutes());
assert_eq!(two_hours_three_min.floor(1.hours() + 5.minutes()), 1.hours() + 5.minutes());pub fn ceil(&self, duration: Duration) -> Duration
pub fn ceil(&self, duration: Duration) -> Duration
Ceils this duration to the closest provided duration
This simply floors then adds the requested duration
Example
use hifitime::{Duration, TimeUnits};
let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.ceil(1.hours()), 3.hours());
assert_eq!(two_hours_three_min.ceil(30.minutes()), 2.hours() + 30.minutes());
assert_eq!(two_hours_three_min.ceil(4.hours()), 4.hours());
assert_eq!(two_hours_three_min.ceil(1.seconds()), two_hours_three_min + 1.seconds());
assert_eq!(two_hours_three_min.ceil(1.hours() + 5.minutes()), 2.hours() + 10.minutes());pub fn round(&self, duration: Duration) -> Duration
pub fn round(&self, duration: Duration) -> Duration
Rounds this duration to the closest provided duration
This performs both a ceil and floor and returns the value which is the closest to current one.
Example
use hifitime::{Duration, TimeUnits};
let two_hours_three_min = 2.hours() + 3.minutes();
assert_eq!(two_hours_three_min.round(1.hours()), 2.hours());
assert_eq!(two_hours_three_min.round(30.minutes()), 2.hours());
assert_eq!(two_hours_three_min.round(4.hours()), 4.hours());
assert_eq!(two_hours_three_min.round(1.seconds()), two_hours_three_min);
assert_eq!(two_hours_three_min.round(1.hours() + 5.minutes()), 2.hours() + 10.minutes());pub fn approx(&self) -> Duration
pub fn approx(&self) -> Duration
Rounds this duration to the largest units represented in this duration.
This is useful to provide an approximate human duration. Under the hood, this function uses round,
so the “tipping point” of the rounding is half way to the next increment of the greatest unit.
As shown below, one example is that 35 hours and 59 minutes rounds to 1 day, but 36 hours and 1 minute rounds
to 2 days because 2 days is closer to 36h 1 min than 36h 1 min is to 1 day.
Example
use hifitime::{Duration, TimeUnits};
assert_eq!((2.hours() + 3.minutes()).approx(), 2.hours());
assert_eq!((24.hours() + 3.minutes()).approx(), 1.days());
assert_eq!((35.hours() + 59.minutes()).approx(), 1.days());
assert_eq!((36.hours() + 1.minutes()).approx(), 2.days());
assert_eq!((47.hours() + 3.minutes()).approx(), 2.days());
assert_eq!((49.hours() + 3.minutes()).approx(), 2.days());pub fn min(self, other: Duration) -> Duration
pub fn min(self, other: Duration) -> Duration
use hifitime::TimeUnits;
let d0 = 20.seconds();
let d1 = 21.seconds();
assert_eq!(d0, d1.min(d0));
assert_eq!(d0, d0.min(d1));pub fn max(self, other: Duration) -> Duration
pub fn max(self, other: Duration) -> Duration
Returns the maximum of the two durations.
use hifitime::TimeUnits;
let d0 = 20.seconds();
let d1 = 21.seconds();
assert_eq!(d1, d1.max(d0));
assert_eq!(d1, d0.max(d1));pub const fn is_negative(&self) -> bool
pub const fn is_negative(&self) -> bool
Returns whether this is a negative or positive duration.