Struct FrameUid
pub struct FrameUid {
pub ephemeris_id: i32,
pub orientation_id: i32,
pub force_inertial: bool,
pub frozen_epoch: Option<Epoch>,
}Expand description
A unique frame reference that only contains enough information to build the actual Frame object. It cannot be used for any computations, is it be used in any structure apart from error structures.
:type ephemeris_id: int :type orientation_id: int
Fields
ephemeris_id: i32orientation_id: i32force_inertial: boolfrozen_epoch: Option<Epoch>Trait Implementations
impl<'de> Deserialize<'de> for FrameUid
impl<'de> Deserialize<'de> for FrameUid
fn deserialize<__D>(
__deserializer: __D,
) -> Result<FrameUid, <__D as Deserializer<'de>>::Error>where
__D: Deserializer<'de>,
fn deserialize<__D>(
__deserializer: __D,
) -> Result<FrameUid, <__D as Deserializer<'de>>::Error>where
__D: Deserializer<'de>,
Deserialize this value from the given Serde deserializer. Read more
impl Encode for FrameUid
impl Encode for FrameUid
fn encoded_len(&self) -> Result<Length, Error>
fn encoded_len(&self) -> Result<Length, Error>
Compute the length of this value in bytes when encoded as ASN.1 DER.
fn encode(&self, encoder: &mut impl Writer) -> Result<(), Error>
fn encode(&self, encoder: &mut impl Writer) -> Result<(), Error>
Encode this value as ASN.1 DER using the provided [
Writer].fn encode_to_slice<'a>(&self, buf: &'a mut [u8]) -> Result<&'a [u8], Error>
fn encode_to_slice<'a>(&self, buf: &'a mut [u8]) -> Result<&'a [u8], Error>
Encode this value to the provided byte slice, returning a sub-slice
containing the encoded message.
fn encode_to_vec(&self, buf: &mut Vec<u8>) -> Result<Length, Error>
fn encode_to_vec(&self, buf: &mut Vec<u8>) -> Result<Length, Error>
Encode this message as ASN.1 DER, appending it to the provided
byte vector.
impl<'py> IntoPyObject<'py> for FrameUid
impl<'py> IntoPyObject<'py> for FrameUid
impl PyTypeInfo for FrameUid
impl PyTypeInfo for FrameUid
const NAME: &'static str = <Self as ::pyo3::PyClass>::NAME
const NAME: &'static str = <Self as ::pyo3::PyClass>::NAME
👎Deprecated since 0.28.0:
prefer using ::type_object(py).name() to get the correct runtime value
Class name.
const MODULE: Option<&'static str> = <Self as ::pyo3::impl_::pyclass::PyClassImpl>::MODULE
const MODULE: Option<&'static str> = <Self as ::pyo3::impl_::pyclass::PyClassImpl>::MODULE
👎Deprecated since 0.28.0:
prefer using ::type_object(py).module() to get the correct runtime value
Module name, if any.
fn type_object_raw(py: Python<'_>) -> *mut PyTypeObject
fn type_object_raw(py: Python<'_>) -> *mut PyTypeObject
Returns the PyTypeObject instance for this type.
fn type_object(py: Python<'_>) -> Bound<'_, PyType>
fn type_object(py: Python<'_>) -> Bound<'_, PyType>
Returns the safe abstraction over the type object.
fn is_type_of(object: &Bound<'_, PyAny>) -> bool
fn is_type_of(object: &Bound<'_, PyAny>) -> bool
Checks if
object is an instance of this type or a subclass of this type.fn is_exact_type_of(object: &Bound<'_, PyAny>) -> bool
fn is_exact_type_of(object: &Bound<'_, PyAny>) -> bool
Checks if
object is an instance of this type.impl Serialize for FrameUid
impl Serialize for FrameUid
fn serialize<__S>(
&self,
__serializer: __S,
) -> Result<<__S as Serializer>::Ok, <__S as Serializer>::Error>where
__S: Serializer,
fn serialize<__S>(
&self,
__serializer: __S,
) -> Result<<__S as Serializer>::Ok, <__S as Serializer>::Error>where
__S: Serializer,
Serialize this value into the given Serde serializer. Read more
impl StaticType for FrameUid
Available on crate feature analysis only.
impl StaticType for FrameUid
Available on crate feature
analysis only.fn static_type() -> SimpleType
fn static_type() -> SimpleType
Return the Dhall type that represents this type. Read more
impl Copy for FrameUid
impl DerefToPyAny for FrameUid
impl Eq for FrameUid
impl StructuralPartialEq for FrameUid