Struct Instrument
pub struct Instrument {
pub q_to_i: EulerParameter,
pub offset_i: Matrix<f64, Const<3>, Const<1>, ArrayStorage<f64, 3, 1>>,
pub fov: FovShape,
}Expand description
Instrument is defined by a mounting Euler Parameter, a mounting translation (“level arm”), and a field of view of the instrument. Notations: frame N is inertial; frame B is body; frame I is instrument.
Fields
q_to_i: EulerParameterThe static rotation from the Parent Frame to the instrument Frame. (How the camera is bolted onto the bus).
offset_i: Matrix<f64, Const<3>, Const<1>, ArrayStorage<f64, 3, 1>>The translation offset from the Parent Frame origin (CoM) to the instrument origin. (The “Lever Arm”).
fov: FovShapeThe geometric definition of the field of view.
Implementations
impl Instrument
impl Instrument
pub fn transform_state(
&self,
q_sc_to_b: EulerParameter,
sc_state: CartesianState,
) -> Result<(EulerParameter, CartesianState), PhysicsError>
pub fn transform_state( &self, q_sc_to_b: EulerParameter, sc_state: CartesianState, ) -> Result<(EulerParameter, CartesianState), PhysicsError>
Computes the state (orientation + Cartesian state) of the instrument at a specific instant, given the spacecraft’s state.
NOTE: This call will return an error if the reference frames are not adequate. Example:
- If the mounting quaterion (q_to_i) frame does not match in sc_attitude_to_body “to” frame IDs
:type q_sc_to_b: Quaternion :type sc_state: Orbit :rtype: tuple[Quaternion, Orbit]
pub fn fov_margin_deg(
&self,
sc_q_to_b: EulerParameter,
sc_state: CartesianState,
target_state: CartesianState,
) -> Result<f64, PhysicsError>
pub fn fov_margin_deg( &self, sc_q_to_b: EulerParameter, sc_state: CartesianState, target_state: CartesianState, ) -> Result<f64, PhysicsError>
Calculates the angular margin to the FOV boundary in degrees.
Arguments
- sc_q_to_b: rotation from the sc_state frame to the body frame in which is expressed the instrument rotation.
- sc_state: state of the spacecraft, typically in an inertial frame
- target_state: state of the target object in the same frame as the sc_state, e.g. IAU Moon if sc_state is in IAU Moon
This is a continuous function suitable for event detection (root finding).
> 0.0: Target is INSIDE.< 0.0: Target is OUTSIDE.= 0.0: Target is ON THE BOUNDARY.
NOTE: This call will return an error if the reference frames are not adequate. Example:
- If the mounting rotation “from” frame does not match in sc_attitude_to_body “to” frame IDs
- If the target state frame ID is not identical to the instrument’s inertial state given the sc_attitude Euler Parameter.
:type sc_q_to_b: Quaternion :type sc_state: Orbit :type target_state: Orbit :rtype: float
pub fn is_target_in_fov(
&self,
sc_attitude_inertial_to_body: EulerParameter,
sc_state: CartesianState,
target_state: CartesianState,
) -> Result<bool, PhysicsError>
pub fn is_target_in_fov( &self, sc_attitude_inertial_to_body: EulerParameter, sc_state: CartesianState, target_state: CartesianState, ) -> Result<bool, PhysicsError>
Checks if a target is visible within the Field of View.
:type sc_attitude_inertial_to_body: Quaternion :type sc_state: Orbit :type target_state: Orbit :rtype: bool
pub fn footprint(
&self,
sc_q_n_to_b: EulerParameter,
sc_state_target: CartesianState,
q_n_to_target: EulerParameter,
resolution: usize,
) -> Result<Vec<CartesianState>, PhysicsError>
pub fn footprint( &self, sc_q_n_to_b: EulerParameter, sc_state_target: CartesianState, q_n_to_target: EulerParameter, resolution: usize, ) -> Result<Vec<CartesianState>, PhysicsError>
Computes the footprint (swath) of the instrument on a target body.
This function projects the edges of the Field of View onto the provided target ellipsoid.
Arguments
sc_q_n_to_b- The orientation of the spacecraft body relative to Inertial.sc_state_target- The inertial state (position/velocity) of the spacecraft.q_n_to_target- The orientation of the target body frame relative to Inertial.resolution- The number of points to generate along the FOV boundary.
Returns
A vector of Orbit objects, each representing a point on the surface of the target
expressed in the target_frame (Fixed).
:type sc_q_n_to_b: Quaternion :type sc_state_target: Orbit :type q_n_to_target: Quaternion :type resolution: int :rtype: list[Orbit]