Struct Polynomial
pub struct Polynomial {
pub constant: Duration,
pub rate: Duration,
pub accel: Duration,
}Expand description
Interpolation Polynomial used for example in [TimeScale] maintenance, precise monitoring or conversions.
Fields
constant: DurationConstant offset Duration, regardless of the interpolation interval
rate: DurationRate or drift in seconds per second (s.s⁻¹) expressed as Duration for convenience. It is a linear scaling factor of the interpolation interval.
accel: DurationAcceleration or drift change in s.s⁻², expressed as Duration for convenience. It is a quadratic scaling of the interpolation interval.
Implementations
impl Polynomial
impl Polynomial
pub fn correction_duration(&self, time_interval: Duration) -> Duration
pub fn correction_duration(&self, time_interval: Duration) -> Duration
impl Polynomial
impl Polynomial
pub fn from_constant_offset(constant: Duration) -> Polynomial
pub fn from_constant_offset(constant: Duration) -> Polynomial
Create a Polynomial structure that is only made of a static offset
pub fn from_constant_offset_nanoseconds(nanos: f64) -> Polynomial
pub fn from_constant_offset_nanoseconds(nanos: f64) -> Polynomial
Create a Polynomial structure from a static offset expressed in nanoseconds
pub fn from_offset_and_rate(constant: Duration, rate: Duration) -> Polynomial
pub fn from_offset_and_rate(constant: Duration, rate: Duration) -> Polynomial
Create a Polynomial structure from both static offset and rate of change:
pub fn from_offset_rate_nanoseconds(
offset_ns: f64,
drift_ns_s: f64,
) -> Polynomial
pub fn from_offset_rate_nanoseconds( offset_ns: f64, drift_ns_s: f64, ) -> Polynomial
Create a Polynomial structure from a static offset and drift, in nanoseconds and nanoseconds.s⁻¹
impl Polynomial
impl Polynomial
pub fn py_from_constant_offset(
_cls: &Bound<'_, PyType>,
constant: Duration,
) -> Polynomial
pub fn py_from_constant_offset( _cls: &Bound<'_, PyType>, constant: Duration, ) -> Polynomial
Create a Polynomial structure that is only made of a static offset :type constant: Duration :rtype: Polynomial
pub fn py_from_constant_offset_nanoseconds(
_cls: &Bound<'_, PyType>,
nanos: f64,
) -> Polynomial
pub fn py_from_constant_offset_nanoseconds( _cls: &Bound<'_, PyType>, nanos: f64, ) -> Polynomial
Create a Polynomial structure from a static offset expressed in nanoseconds :type nanos: float :rtype: Polynomial
pub fn py_from_offset_and_rate(
_cls: &Bound<'_, PyType>,
constant: Duration,
rate: Duration,
) -> Polynomial
pub fn py_from_offset_and_rate( _cls: &Bound<'_, PyType>, constant: Duration, rate: Duration, ) -> Polynomial
Create a Polynomial structure from both static offset and rate of change: :type constant: Duration :type rate: Duration :rtype: Polynomial
pub fn py_from_offset_rate_nanoseconds(
_cls: &Bound<'_, PyType>,
offset_ns: f64,
drift_ns_s: f64,
) -> Polynomial
pub fn py_from_offset_rate_nanoseconds( _cls: &Bound<'_, PyType>, offset_ns: f64, drift_ns_s: f64, ) -> Polynomial
Create a Polynomial structure from a static offset and drift, in nanoseconds and nanoseconds.s⁻¹ :type offset_ns: float :type drift_ns_s: float :rtype: Polynomial