pub struct ThrustDirection {
pub x: f64,
pub y: f64,
pub z: f64,
}Expand description
Applied thrust direction stored in inertial coordinates for the current propagated state.
Fields
x: f64y: f64z: f64Trait Implementations
Sourceimpl Clone for ThrustDirection
impl Clone for ThrustDirection
Sourcefn clone(&self) -> ThrustDirection
fn clone(&self) -> ThrustDirection
Returns a duplicate of the value. Read more
1.0.0 (const: unstable) · Sourcefn clone_from(&mut self, source: &Self)
fn clone_from(&mut self, source: &Self)
Performs copy-assignment from
source. Read moreSourceimpl Debug for ThrustDirection
impl Debug for ThrustDirection
Sourceimpl Default for ThrustDirection
impl Default for ThrustDirection
Sourcefn default() -> ThrustDirection
fn default() -> ThrustDirection
Returns the “default value” for a type. Read more
Sourceimpl<'de> Deserialize<'de> for ThrustDirection
impl<'de> Deserialize<'de> for ThrustDirection
Sourcefn deserialize<__D>(__deserializer: __D) -> Result<Self, __D::Error>where
__D: Deserializer<'de>,
fn deserialize<__D>(__deserializer: __D) -> Result<Self, __D::Error>where
__D: Deserializer<'de>,
Deserialize this value from the given Serde deserializer. Read more
Sourceimpl From<Matrix<f64, Const<3>, Const<1>, ArrayStorage<f64, 3, 1>>> for ThrustDirection
impl From<Matrix<f64, Const<3>, Const<1>, ArrayStorage<f64, 3, 1>>> for ThrustDirection
Sourceimpl From<ThrustDirection> for Vector3<f64>
impl From<ThrustDirection> for Vector3<f64>
Sourcefn from(direction: ThrustDirection) -> Self
fn from(direction: ThrustDirection) -> Self
Converts to this type from the input type.
Sourceimpl<'a, 'py> FromPyObject<'a, 'py> for ThrustDirectionwhere
Self: Clone,
impl<'a, 'py> FromPyObject<'a, 'py> for ThrustDirectionwhere
Self: Clone,
Sourceimpl<'py> IntoPyObject<'py> for ThrustDirection
impl<'py> IntoPyObject<'py> for ThrustDirection
Sourcetype Target = ThrustDirection
type Target = ThrustDirection
The Python output type
Sourcetype Output = Bound<'py, <ThrustDirection as IntoPyObject<'py>>::Target>
type Output = Bound<'py, <ThrustDirection as IntoPyObject<'py>>::Target>
The smart pointer type to use. Read more
Sourcefn into_pyobject(
self,
py: Python<'py>,
) -> Result<<Self as IntoPyObject<'_>>::Output, <Self as IntoPyObject<'_>>::Error>
fn into_pyobject( self, py: Python<'py>, ) -> Result<<Self as IntoPyObject<'_>>::Output, <Self as IntoPyObject<'_>>::Error>
Performs the conversion.
Sourceimpl PartialEq for ThrustDirection
impl PartialEq for ThrustDirection
Sourcefn eq(&self, other: &ThrustDirection) -> bool
fn eq(&self, other: &ThrustDirection) -> bool
Tests for
self and other values to be equal, and is used by ==.Sourceimpl PyClass for ThrustDirection
impl PyClass for ThrustDirection
Sourceimpl PyClassImpl for ThrustDirection
impl PyClassImpl for ThrustDirection
Sourceconst MODULE: Option<&str> = ::core::option::Option::None
const MODULE: Option<&str> = ::core::option::Option::None
Module which the class will be associated with. Read more
Sourceconst IS_BASETYPE: bool = false
const IS_BASETYPE: bool = false
#[pyclass(subclass)]
Sourceconst IS_SUBCLASS: bool = false
const IS_SUBCLASS: bool = false
#[pyclass(extends=…)]
Sourceconst IS_MAPPING: bool = false
const IS_MAPPING: bool = false
#[pyclass(mapping)]
Sourceconst IS_SEQUENCE: bool = false
const IS_SEQUENCE: bool = false
#[pyclass(sequence)]
Sourceconst IS_IMMUTABLE_TYPE: bool = false
const IS_IMMUTABLE_TYPE: bool = false
#[pyclass(immutable_type)]
Sourceconst RAW_DOC: &'static CStr = /// Applied thrust direction stored in inertial coordinates for the current propagated state.
const RAW_DOC: &'static CStr = /// Applied thrust direction stored in inertial coordinates for the current propagated state.
Docstring for the class provided on the struct or enum. Read more
Sourceconst DOC: &'static CStr
const DOC: &'static CStr
Fully rendered class doc, including the
text_signature if a constructor is defined. Read moreSourcetype Layout = <<ThrustDirection as PyClassImpl>::BaseNativeType as PyClassBaseType>::Layout<ThrustDirection>
type Layout = <<ThrustDirection as PyClassImpl>::BaseNativeType as PyClassBaseType>::Layout<ThrustDirection>
Description of how this class is laid out in memory
Sourcetype ThreadChecker = NoopThreadChecker
type ThreadChecker = NoopThreadChecker
This handles following two situations: Read more
type Inventory = Pyo3MethodsInventoryForThrustDirection
Sourcetype PyClassMutability = <<PyAny as PyClassBaseType>::PyClassMutability as PyClassMutability>::MutableChild
type PyClassMutability = <<PyAny as PyClassBaseType>::PyClassMutability as PyClassMutability>::MutableChild
Immutable or mutable
Sourcetype BaseNativeType = PyAny
type BaseNativeType = PyAny
The closest native ancestor. This is
PyAny by default, and when you declare
#[pyclass(extends=PyDict)], it’s PyDict.fn items_iter() -> PyClassItemsIter
fn lazy_type_object() -> &'static LazyTypeObject<Self>
fn dict_offset() -> Option<PyObjectOffset>
fn dict_offset() -> Option<PyObjectOffset>
Used to provide the dictoffset slot
(equivalent to tp_dictoffset)
fn weaklist_offset() -> Option<PyObjectOffset>
fn weaklist_offset() -> Option<PyObjectOffset>
Used to provide the weaklistoffset slot
(equivalent to tp_weaklistoffset
Sourceimpl PyTypeInfo for ThrustDirection
impl PyTypeInfo for ThrustDirection
Sourceconst NAME: &str = <Self as ::pyo3::PyClass>::NAME
const NAME: &str = <Self as ::pyo3::PyClass>::NAME
👎Deprecated since 0.28.0:
prefer using ::type_object(py).name() to get the correct runtime value
Class name.
Sourceconst MODULE: Option<&str> = <Self as ::pyo3::impl_::pyclass::PyClassImpl>::MODULE
const MODULE: Option<&str> = <Self as ::pyo3::impl_::pyclass::PyClassImpl>::MODULE
👎Deprecated since 0.28.0:
prefer using ::type_object(py).module() to get the correct runtime value
Module name, if any.
Sourcefn type_object_raw(py: Python<'_>) -> *mut PyTypeObject
fn type_object_raw(py: Python<'_>) -> *mut PyTypeObject
Returns the PyTypeObject instance for this type.
fn type_object(py: Python<'_>) -> Bound<'_, PyType>
fn type_object(py: Python<'_>) -> Bound<'_, PyType>
Returns the safe abstraction over the type object.
fn is_type_of(object: &Bound<'_, PyAny>) -> bool
fn is_type_of(object: &Bound<'_, PyAny>) -> bool
Checks if
object is an instance of this type or a subclass of this type.fn is_exact_type_of(object: &Bound<'_, PyAny>) -> bool
fn is_exact_type_of(object: &Bound<'_, PyAny>) -> bool
Checks if
object is an instance of this type.Sourceimpl Serialize for ThrustDirection
impl Serialize for ThrustDirection
impl Copy for ThrustDirection
impl DerefToPyAny for ThrustDirection
impl StructuralPartialEq for ThrustDirection