#[repr(u8)]pub enum GuidanceMode {
Coast = 0,
Thrust = 1,
Inhibit = 2,
}Variants
Coast = 0
Guidance is turned off and Guidance Law may switch mode to Thrust for next call
Thrust = 1
Guidance is turned on and Guidance Law may switch mode to Coast for next call
Inhibit = 2
Guidance is turned off and Guidance Law may not change its mode (will need to be done externally to the guidance law).