Spacecraft

Struct Spacecraft 

pub struct Spacecraft {
    pub orbit: Orbit,
    pub mass: Mass,
    pub srp: SRPData,
    pub drag: DragData,
    pub thruster: Option<Thruster>,
    pub mode: GuidanceMode,
    pub thrust_direction: Option<ThrustDirection>,
    pub stm: Option<OMatrix<f64, Const<9>, Const<9>>>,
}
A spacecraft state, composed of its orbit, its masses (dry, prop, extra, all in kg), its SRP configuration, its drag configuration, its thruster configuration, and its guidance mode.

Optionally, the spacecraft state can also store the state transition matrix from the start of the propagation until the current time (i.e. trajectory STM, not step-size STM).

Fields

orbit: Orbit

Initial orbit of the vehicle

mass: Mass

Dry, propellant, and extra masses

srp: SRPData

Solar Radiation Pressure configuration for this spacecraft

drag: DragDatathruster: Option<Thruster>mode: GuidanceMode

Any extra information or extension that is needed for specific guidance laws

thrust_direction: Option<ThrustDirection>

Optionally stores the applied thrust direction for this state in inertial coordinates.

stm: Option<OMatrix<f64, Const<9>, Const<9>>>

Optionally stores the state transition matrix from the start of the propagation until the current time (i.e. trajectory STM, not step-size STM) STM is contains position and velocity, Cr, Cd, prop mass

Implementations

Source

impl Spacecraft

pub fn from_asn1(_cls: &Bound<'_, PyType>, data: &[u8]) -> PyResult<Self>

Decodes an ASN.1 DER encoded byte array into a Mass object.

:type data: bytes :rtype: Mass

pub fn to_asn1<'py>(&self, py: Python<'py>) -> PyResult<Bound<'py, PyBytes>>

Encodes this Mass object into an ASN.1 DER encoded byte array.

:rtype: bytes

Source

impl Spacecraft

pub fn builder() -> SpacecraftBuilder<((), (), (), (), (), (), (), ())>

Create a builder for building Spacecraft. On the builder, call .orbit(...), .mass(...)(optional), .srp(...)(optional), .drag(...)(optional), .thruster(...)(optional), .mode(...)(optional), .thrust_direction(...)(optional), .stm(...)(optional) to set the values of the fields. Finally, call .build() to create the instance of Spacecraft.

Examples found in repository?
nyx-core/examples/03_geo_analysis/raise_optim.rs (line 141)
129fn evaluate_weights(
130    weights: &[f32],
131    prop_time_days: f64,
132    state: Arc<SharedState>,
133) -> Result<(f64, f64), Box<dyn Error>> {
134    let ηthresholds: Vec<f64> = weights.iter().map(|w| *w as f64).collect();
135
136    let eme2k = state.almanac.frame_info(EARTH_J2000).unwrap();
137    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
138
139    let orbit = Orbit::keplerian(24505.9, 0.725, 7.05, 0.0, 0.0, 0.0, epoch, eme2k);
140
141    let sc = Spacecraft::builder()
142        .orbit(orbit)
143        .mass(Mass::from_dry_and_prop_masses(1000.0, 1000.0))
144        .srp(SRPData::from_area(3.0 * 6.0))
145        .thruster(Thruster {
146            isp_s: 4435.0,
147            thrust_N: 0.472,
148        })
149        .mode(GuidanceMode::Thrust)
150        .build();
151
152    let prop_time = prop_time_days * Unit::Day;
153
154    let objectives = &[
155        Objective::within_tolerance(
156            StateParameter::Element(OrbitalElement::SemiMajorAxis),
157            30_000.0,
158            20.0,
159        ),
160        Objective::within_tolerance(
161            StateParameter::Element(OrbitalElement::Eccentricity),
162            0.001,
163            5e-5,
164        ),
165        Objective::within_tolerance(
166            StateParameter::Element(OrbitalElement::Inclination),
167            0.05,
168            1e-2,
169        ),
170    ];
171
172    // let kluever_ctrl = Kluever::from_max_eclipse(objectives, &weights_f64, 0.2);
173    let ctrl = Ruggiero::from_ηthresholds(objectives, &ηthresholds, sc)?;
174
175    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
176    orbital_dyn.accel_models.push(state.harmonics.clone());
177
178    let sc_dynamics = SpacecraftDynamics::from_model(orbital_dyn, state.srp_dyn.clone())
179        .with_guidance_law(ctrl.clone());
180
181    let (final_state, _traj) = Propagator::rk89(
182        sc_dynamics.clone(),
183        IntegratorOptions::builder()
184            .min_step(10.0_f64.seconds())
185            .tolerance(1e-8)
186            .error_ctrl(ErrorControl::RSSCartesianStep)
187            .build(),
188    )
189    .with(sc, state.almanac.clone())
190    .for_duration_with_traj(prop_time)?;
191
192    let prop_usage = sc.mass.prop_mass_kg - final_state.mass.prop_mass_kg;
193
194    let mut penalty = 0.0;
195    for obj in objectives {
196        let (achieved, error) = obj.assess(&final_state)?;
197        if !achieved {
198            penalty += error.abs();
199        }
200        info!("{obj} error: {error:.3}, achieved? {achieved}");
201    }
202
203    info!("{ηthresholds:?} -> {prop_usage:.3} kg\tpenalty = {penalty:.3}");
204
205    Ok((prop_usage, penalty * 1000.0))
206}
More examples
Hide additional examples
nyx-core/examples/03_geo_analysis/stationkeeping.rs (line 39)
28fn main() -> Result<(), Box<dyn Error>> {
29    pel::init();
30    // Set up the dynamics like in the orbit raise.
31    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
32    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
33
34    // Define the GEO orbit, and we're just going to maintain it very tightly.
35    let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
36    let orbit = Orbit::try_keplerian(42164.0, 1e-5, 0., 163.0, 75.0, 0.0, epoch, earth_j2000)?;
37    println!("{orbit:x}");
38
39    let sc = Spacecraft::builder()
40        .orbit(orbit)
41        .mass(Mass::from_dry_and_prop_masses(1000.0, 1000.0)) // 1000 kg of dry mass and prop, totalling 2.0 tons
42        .srp(SRPData::from_area(3.0 * 6.0)) // Assuming 1 kW/m^2 or 18 kW, giving a margin of 4.35 kW for on-propulsion consumption
43        .thruster(Thruster {
44            // "NEXT-STEP" row in Table 2
45            isp_s: 4435.0,
46            thrust_N: 0.472,
47        })
48        .mode(GuidanceMode::Thrust) // Start thrusting immediately.
49        .build();
50
51    // Set up the spacecraft dynamics like in the orbit raise example.
52
53    let prop_time = 30.0 * Unit::Day;
54
55    // Define the guidance law -- we're just using a Ruggiero controller as demonstrated in AAS-2004-5089.
56    let objectives = &[
57        Objective::within_tolerance(
58            StateParameter::Element(OrbitalElement::SemiMajorAxis),
59            42_165.0,
60            20.0,
61        ),
62        Objective::within_tolerance(
63            StateParameter::Element(OrbitalElement::Eccentricity),
64            0.001,
65            5e-5,
66        ),
67        Objective::within_tolerance(
68            StateParameter::Element(OrbitalElement::Inclination),
69            0.05,
70            1e-2,
71        ),
72    ];
73
74    let ruggiero_ctrl = Ruggiero::from_max_eclipse(objectives, sc, 0.2)?;
75    println!("{ruggiero_ctrl}");
76
77    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
78
79    let mut jgm3_meta = MetaFile {
80        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
81        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
82    };
83    jgm3_meta.process(true)?;
84
85    let harmonics = GravityField::from_stor(
86        almanac.frame_info(IAU_EARTH_FRAME)?,
87        GravityFieldData::from_cof(&jgm3_meta.uri, 8, 8, true)?,
88    );
89    orbital_dyn.accel_models.push(harmonics);
90
91    let srp_dyn = SolarPressure::default_flux(EARTH_J2000, almanac.clone())?;
92    let sc_dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn)
93        .with_guidance_law(ruggiero_ctrl.clone());
94
95    println!("{sc_dynamics}");
96
97    // Finally, let's use the Monte Carlo framework built into Nyx to propagate spacecraft.
98
99    // Let's start by defining the dispersion.
100    // The MultivariateNormal structure allows us to define the dispersions in any of the orbital parameters, but these are applied directly in the Cartesian state space.
101    // Note that additional validation on the MVN is in progress -- https://github.com/nyx-space/nyx/issues/339.
102    let mc_rv = MvnSpacecraft::new(
103        sc,
104        vec![StateDispersion::zero_mean(
105            StateParameter::Element(OrbitalElement::SemiMajorAxis),
106            3.0,
107        )],
108    )?;
109
110    let my_mc = MonteCarlo::new(
111        sc, // Nominal state
112        mc_rv,
113        "03_geo_sk".to_string(), // Scenario name
114        None, // No specific seed specified, so one will be drawn from the computer's entropy.
115    );
116
117    // Build the propagator setup.
118    let setup = Propagator::rk89(
119        sc_dynamics.clone(),
120        IntegratorOptions::builder()
121            .min_step(10.0_f64.seconds())
122            .error_ctrl(ErrorControl::RSSCartesianStep)
123            .build(),
124    );
125
126    let num_runs = 25;
127    let rslts = my_mc.run_until_epoch(setup, almanac.clone(), sc.epoch() + prop_time, num_runs);
128
129    assert_eq!(rslts.runs.len(), num_runs);
130
131    rslts.to_parquet("03_geo_sk.parquet", ExportCfg::default())?;
132
133    Ok(())
134}
nyx-core/examples/03_geo_analysis/raise.rs (line 52)
27fn main() -> Result<(), Box<dyn Error>> {
28    pel::init();
29
30    // Dynamics models require planetary constants and ephemerides to be defined.
31    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
32    // This will automatically download the DE440s planetary ephemeris,
33    // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
34    // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
35    // planetary constants kernels.
36    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
37    // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
38    // references to many functions.
39    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
40    // Fetch the EME2000 frame from the Almabac
41    let eme2k = almanac.frame_info(EARTH_J2000).unwrap();
42    // Define the orbit epoch
43    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
44
45    // Build the spacecraft itself.
46    // Using slide 6 of https://aerospace.org/sites/default/files/2018-11/Davis-Mayberry_HPSEP_11212018.pdf
47    // for the "next gen" SEP characteristics.
48
49    // GTO start
50    let orbit = Orbit::keplerian(24505.9, 0.725, 7.05, 0.0, 0.0, 0.0, epoch, eme2k);
51
52    let sc = Spacecraft::builder()
53        .orbit(orbit)
54        .mass(Mass::from_dry_and_prop_masses(1000.0, 1000.0)) // 1000 kg of dry mass and prop, totalling 2.0 tons
55        .srp(SRPData::from_area(3.0 * 6.0)) // Assuming 1 kW/m^2 or 18 kW, giving a margin of 4.35 kW for on-propulsion consumption
56        .thruster(Thruster {
57            // "NEXT-STEP" row in Table 2
58            isp_s: 4435.0,
59            thrust_N: 0.472,
60        })
61        .mode(GuidanceMode::Thrust) // Start thrusting immediately.
62        .build();
63
64    let prop_time = 180.0 * Unit::Day;
65
66    // Define the guidance law -- we're just using a Ruggiero controller as demonstrated in AAS-2004-5089.
67    let objectives = &[
68        Objective::within_tolerance(
69            StateParameter::Element(OrbitalElement::SemiMajorAxis),
70            42_165.0,
71            20.0,
72        ),
73        Objective::within_tolerance(
74            StateParameter::Element(OrbitalElement::Eccentricity),
75            0.001,
76            5e-5,
77        ),
78        Objective::within_tolerance(
79            StateParameter::Element(OrbitalElement::Inclination),
80            0.05,
81            1e-2,
82        ),
83    ];
84
85    // Ensure that we only thrust if we have more than 20% illumination.
86    let ruggiero_ctrl = Ruggiero::from_max_eclipse(objectives, sc, 0.2).unwrap();
87    println!("{ruggiero_ctrl}");
88
89    // Define the high fidelity dynamics
90
91    // Set up the spacecraft dynamics.
92
93    // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
94    // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
95    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
96
97    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
98    // We're using the JGM3 model here, which is the default in GMAT.
99    let mut jgm3_meta = MetaFile {
100        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
101        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
102    };
103    // And let's download it if we don't have it yet.
104    jgm3_meta.process(true)?;
105
106    // Build the spherical harmonics.
107    // The harmonics must be computed in the body fixed frame.
108    // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
109    let harmonics = GravityField::from_stor(
110        almanac.frame_info(IAU_EARTH_FRAME)?,
111        GravityFieldData::from_cof(&jgm3_meta.uri, 8, 8, true).unwrap(),
112    );
113
114    // Include the spherical harmonics into the orbital dynamics.
115    orbital_dyn.accel_models.push(harmonics);
116
117    // We define the solar radiation pressure, using the default solar flux and accounting only
118    // for the eclipsing caused by the Earth.
119    let srp_dyn = SolarPressure::default_flux(EARTH_J2000, almanac.clone())?;
120
121    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
122    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
123    let sc_dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn)
124        .with_guidance_law(ruggiero_ctrl.clone());
125
126    println!("{orbit:x}");
127
128    // We specify a minimum step in the propagator because the Ruggiero control would otherwise drive this step very low.
129    let (final_state, traj) = Propagator::rk89(
130        sc_dynamics.clone(),
131        IntegratorOptions::builder()
132            .min_step(10.0_f64.seconds())
133            .error_ctrl(ErrorControl::RSSCartesianStep)
134            .build(),
135    )
136    .with(sc, almanac.clone())
137    .for_duration_with_traj(prop_time)?;
138
139    let prop_usage = sc.mass.prop_mass_kg - final_state.mass.prop_mass_kg;
140    println!("{:x}", final_state.orbit);
141    println!("prop usage: {prop_usage:.3} kg");
142
143    // Finally, export the results for analysis, including the penumbra percentage throughout the orbit raise.
144    traj.to_parquet("./03_geo_raise.parquet", ExportCfg::default())?;
145
146    for status_line in ruggiero_ctrl.status(&final_state) {
147        println!("{status_line}");
148    }
149
150    ruggiero_ctrl
151        .achieved(&final_state)
152        .expect("objective not achieved");
153
154    Ok(())
155}
nyx-core/examples/02_jwst_covar_monte_carlo/main.rs (line 63)
26fn main() -> Result<(), Box<dyn Error>> {
27    pel::init();
28    // Dynamics models require planetary constants and ephemerides to be defined.
29    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
30    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
31
32    // Download the regularly update of the James Webb Space Telescope reconstucted (or definitive) ephemeris.
33    // Refer to https://naif.jpl.nasa.gov/pub/naif/JWST/kernels/spk/aareadme.txt for details.
34    let mut latest_jwst_ephem = MetaFile {
35        uri: "https://naif.jpl.nasa.gov/pub/naif/JWST/kernels/spk/jwst_rec.bsp".to_string(),
36        crc32: None,
37    };
38    latest_jwst_ephem.process(true)?;
39
40    // Load this ephem in the general Almanac we're using for this analysis.
41    let almanac = Arc::new(
42        MetaAlmanac::latest()
43            .map_err(Box::new)?
44            .load_from_metafile(latest_jwst_ephem, true)?,
45    );
46
47    // By loading this ephemeris file in the ANISE GUI or ANISE CLI, we can find the NAIF ID of the JWST
48    // in the BSP. We need this ID in order to query the ephemeris.
49    const JWST_NAIF_ID: i32 = -170;
50    // Let's build a frame in the J2000 orientation centered on the JWST.
51    const JWST_J2000: Frame = Frame::from_ephem_j2000(JWST_NAIF_ID);
52
53    // Since the ephemeris file is updated regularly, we'll just grab the latest state in the ephem.
54    let (earliest_epoch, latest_epoch) = almanac.spk_domain(JWST_NAIF_ID)?;
55    println!("JWST defined from {earliest_epoch} to {latest_epoch}");
56    // Fetch the state, printing it in the Earth J2000 frame.
57    let jwst_orbit = almanac.transform(JWST_J2000, EARTH_J2000, latest_epoch, None)?;
58    println!("{jwst_orbit:x}");
59
60    // Build the spacecraft
61    // SRP area assumed to be the full sunshield and mass if 6200.0 kg, c.f. https://webb.nasa.gov/content/about/faqs/facts.html
62    // SRP Coefficient of reflectivity assumed to be that of Kapton, i.e. 2 - 0.44 = 1.56, table 1 from https://amostech.com/TechnicalPapers/2018/Poster/Bengtson.pdf
63    let jwst = Spacecraft::builder()
64        .orbit(jwst_orbit)
65        .srp(SRPData {
66            area_m2: 21.197 * 14.162,
67            coeff_reflectivity: 1.56,
68        })
69        .mass(Mass::from_dry_mass(6200.0))
70        .build();
71
72    // Build up the spacecraft uncertainty builder.
73    // We can use the spacecraft uncertainty structure to build this up.
74    // We start by specifying the nominal state (as defined above), then the uncertainty in position and velocity
75    // in the RIC frame. We could also specify the Cr, Cd, and mass uncertainties, but these aren't accounted for until
76    // Nyx can also estimate the deviation of the spacecraft parameters.
77    let jwst_uncertainty = SpacecraftUncertainty::builder()
78        .nominal(jwst)
79        .frame(LocalFrame::RIC)
80        .x_km(0.5)
81        .y_km(0.3)
82        .z_km(1.5)
83        .vx_km_s(1e-4)
84        .vy_km_s(0.6e-3)
85        .vz_km_s(3e-3)
86        .build();
87
88    println!("{jwst_uncertainty}");
89
90    // Build the Kalman filter estimate.
91    // Note that we could have used the KfEstimate structure directly (as seen throughout the OD integration tests)
92    // but this approach requires quite a bit more boilerplate code.
93    let jwst_estimate = jwst_uncertainty.to_estimate()?;
94
95    // Set up the spacecraft dynamics.
96    // We'll use the point masses of the Earth, Sun, Jupiter (barycenter, because it's in the DE440), and the Moon.
97    // We'll also enable solar radiation pressure since the James Webb has a huge and highly reflective sun shield.
98
99    let orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN, JUPITER_BARYCENTER]);
100    let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], almanac.clone())?;
101
102    // Finalize setting up the dynamics.
103    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
104
105    // Build the propagator set up to use for the whole analysis.
106    let setup = Propagator::default(dynamics);
107
108    // All of the analysis will use this duration.
109    let prediction_duration = 6.5 * Unit::Day;
110
111    // === Covariance mapping ===
112    // For the covariance mapping / prediction, we'll use the common orbit determination approach.
113    // This is done by setting up a spacecraft Kalman filter OD process, and predicting for the analysis duration.
114
115    // Build the propagation instance for the OD process.
116    let odp = SpacecraftKalmanOD::new(
117        setup.clone(),
118        KalmanVariant::DeviationTracking,
119        None,
120        BTreeMap::new(),
121        almanac.clone(),
122    );
123
124    // The prediction step is 1 minute by default, configured in the OD process, i.e. how often we want to know the covariance.
125    assert_eq!(odp.max_step, 1_i64.minutes());
126    // Finally, predict, and export the trajectory with covariance to a parquet file.
127    let od_sol = odp.predict_for(jwst_estimate, prediction_duration)?;
128    od_sol.to_parquet("./02_jwst_covar_map.parquet", ExportCfg::default())?;
129
130    // === Monte Carlo framework ===
131    // Nyx comes with a complete multi-threaded Monte Carlo frame. It's blazing fast.
132
133    let my_mc = MonteCarlo::new(
134        jwst, // Nominal state
135        jwst_estimate.to_random_variable()?,
136        "02_jwst".to_string(), // Scenario name
137        None, // No specific seed specified, so one will be drawn from the computer's entropy.
138    );
139
140    let num_runs = 5_000;
141    let rslts = my_mc.run_until_epoch(
142        setup,
143        almanac.clone(),
144        jwst.epoch() + prediction_duration,
145        num_runs,
146    );
147
148    assert_eq!(rslts.runs.len(), num_runs);
149    // Finally, export these results, computing the eclipse percentage for all of these results.
150
151    rslts.to_parquet("02_jwst_monte_carlo.parquet", ExportCfg::default())?;
152
153    Ok(())
154}
nyx-core/examples/05_cislunar_spacecraft_link_od/main.rs (line 104)
34fn main() -> Result<(), Box<dyn Error>> {
35    pel::init();
36
37    // ====================== //
38    // === ALMANAC SET UP === //
39    // ====================== //
40
41    let manifest_dir = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
42
43    let out = manifest_dir.join("data/04_output/");
44
45    let almanac = Arc::new(
46        Almanac::new(
47            &manifest_dir
48                .join("data/01_planetary/pck08.pca")
49                .to_string_lossy(),
50        )
51        .unwrap()
52        .load(
53            &manifest_dir
54                .join("data/01_planetary/de440s.bsp")
55                .to_string_lossy(),
56        )
57        .unwrap(),
58    );
59
60    let eme2k = almanac.frame_info(EARTH_J2000).unwrap();
61    let moon_iau = almanac.frame_info(IAU_MOON_FRAME).unwrap();
62
63    let epoch = Epoch::from_gregorian_tai(2021, 5, 29, 19, 51, 16, 852_000);
64    let nrho = Orbit::cartesian(
65        166_473.631_302_239_7,
66        -274_715.487_253_382_7,
67        -211_233.210_176_686_7,
68        0.933_451_604_520_018_4,
69        0.436_775_046_841_900_9,
70        -0.082_211_021_250_348_95,
71        epoch,
72        eme2k,
73    );
74
75    let tx_nrho_sc = Spacecraft::from(nrho);
76
77    let state_luna = almanac.transform_to(nrho, MOON_J2000, None).unwrap();
78    println!("Start state (dynamics: Earth, Moon, Sun gravity):\n{state_luna}");
79
80    let bodies = vec![EARTH, SUN];
81    let dynamics = SpacecraftDynamics::new(OrbitalDynamics::point_masses(bodies));
82
83    let setup = Propagator::rk89(
84        dynamics,
85        IntegratorOptions::builder().max_step(0.5.minutes()).build(),
86    );
87
88    /* == Propagate the NRHO vehicle == */
89    let prop_time = 1.1 * state_luna.period().unwrap();
90
91    let (nrho_final, mut tx_traj) = setup
92        .with(tx_nrho_sc, almanac.clone())
93        .for_duration_with_traj(prop_time)
94        .unwrap();
95
96    tx_traj.name = Some("NRHO Tx SC".to_string());
97
98    println!("{tx_traj}");
99
100    /* == Propagate an LLO vehicle == */
101    let llo_orbit =
102        Orbit::try_keplerian_altitude(110.0, 1e-4, 90.0, 0.0, 0.0, 0.0, epoch, moon_iau).unwrap();
103
104    let llo_sc = Spacecraft::builder().orbit(llo_orbit).build();
105
106    let (_, llo_traj) = setup
107        .with(llo_sc, almanac.clone())
108        .until_epoch_with_traj(nrho_final.epoch())
109        .unwrap();
110
111    // Export the subset of the first two hours.
112    llo_traj
113        .clone()
114        .filter_by_offset(..2.hours())
115        .to_parquet_simple(out.join("05_caps_llo_truth.pq"))?;
116
117    /* == Setup the interlink == */
118
119    let mut measurement_types = IndexSet::new();
120    measurement_types.insert(MeasurementType::Range);
121    measurement_types.insert(MeasurementType::Doppler);
122
123    let mut stochastics = IndexMap::new();
124
125    let sa45_csac_allan_dev = 1e-11;
126
127    stochastics.insert(
128        MeasurementType::Range,
129        StochasticNoise::from_hardware_range_km(
130            sa45_csac_allan_dev,
131            10.0.seconds(),
132            link_specific::ChipRate::StandardT4B,
133            link_specific::SN0::Average,
134        ),
135    );
136
137    stochastics.insert(
138        MeasurementType::Doppler,
139        StochasticNoise::from_hardware_doppler_km_s(
140            sa45_csac_allan_dev,
141            10.0.seconds(),
142            link_specific::CarrierFreq::SBand,
143            link_specific::CN0::Average,
144        ),
145    );
146
147    let interlink = InterlinkTxSpacecraft {
148        traj: tx_traj,
149        measurement_types,
150        integration_time: None,
151        timestamp_noise_s: None,
152        ab_corr: Aberration::LT,
153        stochastic_noises: Some(stochastics),
154    };
155
156    // Devices are the transmitter, which is our NRHO vehicle.
157    let mut devices = BTreeMap::new();
158    devices.insert("NRHO Tx SC".to_string(), interlink);
159
160    let mut configs = BTreeMap::new();
161    configs.insert(
162        "NRHO Tx SC".to_string(),
163        TrkConfig::builder()
164            .strands(vec![Strand {
165                start: epoch,
166                end: nrho_final.epoch(),
167            }])
168            .build(),
169    );
170
171    let mut trk_sim =
172        TrackingArcSim::with_seed(devices.clone(), llo_traj.clone(), configs, 0).unwrap();
173    println!("{trk_sim}");
174
175    let trk_data = trk_sim.generate_measurements(almanac.clone()).unwrap();
176    println!("{trk_data}");
177
178    trk_data
179        .to_parquet_simple(out.clone().join("nrho_interlink_msr.pq"))
180        .unwrap();
181
182    // Run a truth OD where we estimate the LLO position
183    let llo_uncertainty = SpacecraftUncertainty::builder()
184        .nominal(llo_sc)
185        .x_km(1.0)
186        .y_km(1.0)
187        .z_km(1.0)
188        .vx_km_s(1e-3)
189        .vy_km_s(1e-3)
190        .vz_km_s(1e-3)
191        .build();
192
193    let mut proc_devices = devices.clone();
194
195    // Define the initial estimate, randomized, seed for reproducibility
196    let mut initial_estimate = llo_uncertainty.to_estimate_randomized(Some(0)).unwrap();
197    // Inflate the covariance -- https://github.com/nyx-space/nyx/issues/339
198    initial_estimate.covar *= 2.5;
199
200    // Increase the noise in the devices to accept more measurements.
201
202    for link in proc_devices.values_mut() {
203        for noise in &mut link.stochastic_noises.as_mut().unwrap().values_mut() {
204            *noise.white_noise.as_mut().unwrap() *= 3.0;
205        }
206    }
207
208    let init_err = initial_estimate
209        .orbital_state()
210        .ric_difference(&llo_orbit)
211        .unwrap();
212
213    println!("initial estimate:\n{initial_estimate}");
214    println!("RIC errors = {init_err}",);
215
216    let odp = InterlinkKalmanOD::new(
217        setup.clone(),
218        KalmanVariant::ReferenceUpdate,
219        Some(ResidRejectCrit::default()),
220        proc_devices,
221        almanac.clone(),
222    );
223
224    // Shrink the data to process.
225    let arc = trk_data.filter_by_offset(..2.hours());
226
227    let od_sol = odp.process_arc(initial_estimate, &arc).unwrap();
228
229    println!("{od_sol}");
230
231    od_sol
232        .to_parquet(
233            out.join("05_caps_interlink_od_sol.pq"),
234            ExportCfg::default(),
235        )
236        .unwrap();
237
238    let od_traj = od_sol.to_traj().unwrap();
239
240    od_traj
241        .ric_diff_to_parquet(
242            &llo_traj,
243            out.join("05_caps_interlink_llo_est_error.pq"),
244            ExportCfg::default(),
245        )
246        .unwrap();
247
248    let final_est = od_sol.estimates.last().unwrap();
249    assert!(final_est.within_3sigma(), "should be within 3 sigma");
250
251    println!("ESTIMATE\n{final_est:x}\n");
252    let truth = llo_traj.at(final_est.epoch()).unwrap();
253    println!("TRUTH\n{truth:x}");
254
255    let final_err = truth
256        .orbit
257        .ric_difference(&final_est.orbital_state())
258        .unwrap();
259    println!("ERROR {final_err}");
260
261    // Build the residuals versus reference plot.
262    let rvr_sol = odp
263        .process_arc(initial_estimate, &arc.resid_vs_ref_check())
264        .unwrap();
265
266    rvr_sol
267        .to_parquet(
268            out.join("05_caps_interlink_resid_v_ref.pq"),
269            ExportCfg::default(),
270        )
271        .unwrap();
272
273    let final_rvr = rvr_sol.estimates.last().unwrap();
274
275    println!("RMAG error {:.3} m", final_err.rmag_km() * 1e3);
276    println!(
277        "Pure prop error {:.3} m",
278        final_rvr
279            .orbital_state()
280            .ric_difference(&final_est.orbital_state())
281            .unwrap()
282            .rmag_km()
283            * 1e3
284    );
285
286    Ok(())
287}
nyx-core/examples/06_lunar_orbit_determination/main.rs (line 71)
35fn main() -> Result<(), Box<dyn Error>> {
36    pel::init();
37
38    // ====================== //
39    // === ALMANAC SET UP === //
40    // ====================== //
41
42    // Dynamics models require planetary constants and ephemerides to be defined.
43    // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45    let data_folder: PathBuf = [
46        env!("CARGO_MANIFEST_DIR"),
47        "examples",
48        "06_lunar_orbit_determination",
49    ]
50    .iter()
51    .collect();
52
53    let meta = data_folder.join("metaalmanac.dhall");
54
55    // Load this ephem in the general Almanac we're using for this analysis.
56    let almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
57        .map_err(Box::new)?
58        .process(true)
59        .map_err(Box::new)?;
60
61    // Lock the almanac (an Arc is a read only structure).
62    let almanac = Arc::new(almanac);
63
64    // Build a nominal trajectory
65    // TODO: Switch this to a sequence once the OD over a spacecraft sequence is implemented.
66
67    let epoch = Epoch::from_gregorian_utc_at_noon(2024, 2, 29);
68    let moon_j2000 = almanac.frame_info(MOON_J2000)?;
69
70    // To build the trajectory we need to provide a spacecraft template.
71    let orbiter = Spacecraft::builder()
72        .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0))
73        .srp(SRPData {
74            area_m2: 3.9 * 2.7,
75            coeff_reflectivity: 0.96,
76        })
77        .orbit(Orbit::try_keplerian_altitude(
78            150.0, 0.00212, 33.6, 45.0, 45.0, 0.0, epoch, moon_j2000,
79        )?) // Setting a zero orbit here because it's just a template
80        .build();
81
82    // ========================== //
83    // === BUILD NOMINAL TRAJ === //
84    // ========================== //
85
86    // Set up the spacecraft dynamics.
87
88    // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
89    // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
90    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
91
92    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
93    // We're using the GRAIL JGGRX model.
94    let mut jggrx_meta = MetaFile {
95        uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
96        crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
97    };
98    // And let's download it if we don't have it yet.
99    jggrx_meta.process(true)?;
100
101    // Build the spherical harmonics.
102    // The harmonics must be computed in the body fixed frame.
103    // We're using the long term prediction of the Moon principal axes frame.
104    let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
105    let sph_harmonics = GravityField::from_stor(
106        almanac.frame_info(moon_pa_frame)?,
107        GravityFieldData::from_shadr(&jggrx_meta.uri, 80, 80, true)?,
108    );
109
110    // Include the spherical harmonics into the orbital dynamics.
111    orbital_dyn.accel_models.push(sph_harmonics);
112
113    // We define the solar radiation pressure, using the default solar flux and accounting only
114    // for the eclipsing caused by the Earth and Moon.
115    // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
116    let srp_dyn = SolarPressure::new(vec![MOON_J2000], almanac.clone())?;
117
118    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
119    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
120    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
121
122    println!("{dynamics}");
123
124    let setup = Propagator::rk89(dynamics.clone(), IntegratorOptions::default());
125
126    let truth_traj = setup
127        .with(orbiter, almanac.clone())
128        .for_duration_with_traj(Unit::Day * 2)?
129        .1;
130
131    // ==================== //
132    // === OD SIMULATOR === //
133    // ==================== //
134
135    // Load the Deep Space Network ground stations.
136    // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
137    let ground_station_file = data_folder.join("dsn-network.yaml");
138    let devices = GroundStation::load_named(ground_station_file)?;
139
140    let proc_devices = devices.clone();
141
142    // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
143    // Nyx can build a tracking schedule for you based on the first station with access.
144    let configs: BTreeMap<String, TrkConfig> =
145        TrkConfig::load_named(data_folder.join("tracking-cfg.yaml"))?;
146
147    // Build the tracking arc simulation to generate a "standard measurement".
148    let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
149        devices.clone(),
150        truth_traj.clone(),
151        configs,
152        123, // Set a seed for reproducibility
153    )?;
154
155    trk.build_schedule(almanac.clone())?;
156    let arc = trk.generate_measurements(almanac.clone())?;
157    // Save the simulated tracking data
158    arc.to_parquet_simple("./data/04_output/06_lunar_simulated_tracking.parquet")?;
159
160    // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
161    println!("{arc}");
162
163    // Now that we have simulated measurements, we'll run the orbit determination.
164
165    // ===================== //
166    // === OD ESTIMATION === //
167    // ===================== //
168
169    let sc = SpacecraftUncertainty::builder()
170        .nominal(orbiter)
171        .frame(LocalFrame::RIC)
172        .x_km(0.5)
173        .y_km(0.5)
174        .z_km(0.5)
175        .vx_km_s(5e-3)
176        .vy_km_s(5e-3)
177        .vz_km_s(5e-3)
178        .build();
179
180    // Build the filter initial estimate, which we will reuse in the filter.
181    let initial_estimate = sc.to_estimate()?;
182
183    println!("== FILTER STATE ==\n{orbiter:x}\n{initial_estimate}");
184
185    // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
186    let process_noise = ProcessNoise3D::from_velocity_km_s(
187        &[1e-14, 1e-14, 1e-14],
188        1 * Unit::Hour,
189        10 * Unit::Minute,
190        None,
191    );
192
193    println!("{process_noise}");
194
195    // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
196    let odp = SpacecraftKalmanScalarOD::new(
197        setup,
198        KalmanVariant::ReferenceUpdate,
199        Some(ResidRejectCrit::default()),
200        proc_devices,
201        almanac.clone(),
202    )
203    .with_process_noise(process_noise);
204
205    let od_sol = odp.process_arc(initial_estimate, &arc)?;
206
207    let final_est = od_sol.estimates.last().unwrap();
208
209    println!("{final_est}");
210
211    let ric_err = truth_traj
212        .at(final_est.epoch())?
213        .orbit
214        .ric_difference(&final_est.orbital_state())?;
215    println!("== RIC at end ==");
216    println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
217    println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
218
219    println!(
220        "Num residuals rejected: #{}",
221        od_sol.rejected_residuals().len()
222    );
223    println!(
224        "Percentage within +/-3: {}",
225        od_sol.residual_ratio_within_threshold(3.0).unwrap()
226    );
227    println!("Whitened residuals normal? {}", od_sol.is_normal(None)?);
228    println!("NIS test success? {}", od_sol.is_nis_consistent(None)?);
229
230    od_sol.to_parquet(
231        "./data/04_output/06_lunar_od_results.parquet",
232        ExportCfg::default(),
233    )?;
234
235    let od_trajectory = od_sol.to_traj()?;
236    // Build the RIC difference.
237    od_trajectory.ric_diff_to_parquet(
238        &truth_traj,
239        "./data/04_output/06_lunar_od_truth_error.parquet",
240        ExportCfg::default(),
241    )?;
242
243    Ok(())
244}
Source

impl Spacecraft

pub fn new( orbit: Orbit, dry_mass_kg: f64, prop_mass_kg: f64, srp_area_m2: f64, drag_area_m2: f64, coeff_reflectivity: f64, coeff_drag: f64, ) -> Self

Initialize a spacecraft state from all of its parameters

pub fn from_thruster( orbit: Orbit, dry_mass_kg: f64, prop_mass_kg: f64, thruster: Thruster, mode: GuidanceMode, ) -> Self

Initialize a spacecraft state from only a thruster and mass. Use this when designing guidance laws while ignoring drag and SRP.

pub fn from_srp_defaults( orbit: Orbit, dry_mass_kg: f64, srp_area_m2: f64, ) -> Self

Initialize a spacecraft state from the SRP default 1.8 for coefficient of reflectivity (prop mass and drag parameters nullified!)

pub fn from_drag_defaults( orbit: Orbit, dry_mass_kg: f64, drag_area_m2: f64, ) -> Self

Initialize a spacecraft state from the SRP default 1.8 for coefficient of drag (prop mass and SRP parameters nullified!)

pub fn with_dv_km_s(self, dv_km_s: Vector3<f64>) -> Self

pub fn with_dry_mass(self, dry_mass_kg: f64) -> Self

Returns a copy of the state with a new dry mass

pub fn with_prop_mass(self, prop_mass_kg: f64) -> Self

Returns a copy of the state with a new prop mass

pub fn with_srp(self, srp_area_m2: f64, coeff_reflectivity: f64) -> Self

Returns a copy of the state with a new SRP area and CR

pub fn with_srp_area(self, srp_area_m2: f64) -> Self

Returns a copy of the state with a new SRP area

pub fn with_cr(self, coeff_reflectivity: f64) -> Self

Returns a copy of the state with a new coefficient of reflectivity

pub fn with_drag(self, drag_area_m2: f64, coeff_drag: f64) -> Self

Returns a copy of the state with a new drag area and CD

pub fn with_drag_area(self, drag_area_m2: f64) -> Self

Returns a copy of the state with a new SRP area

pub fn with_cd(self, coeff_drag: f64) -> Self

Returns a copy of the state with a new coefficient of drag

pub fn with_orbit(self, orbit: Orbit) -> Self

Returns a copy of the state with a new orbit

pub fn enable_stm(&mut self)

Sets the STM of this state of identity, which also enables computation of the STM for spacecraft navigation

pub fn mass_kg(&self) -> f64

Returns the total mass in kilograms

pub fn with_guidance_mode(self, mode: GuidanceMode) -> Self

Returns a copy of the state with the provided guidance mode

pub fn mode(&self) -> GuidanceMode

pub fn mut_mode(&mut self, mode: GuidanceMode)

pub fn thrust_direction(&self) -> Option<Vector3<f64>>

Return the thrust direction, if there one a predefined one, as a unit vector.

pub fn mut_thrust_direction(&mut self, direction: Option<Vector3<f64>>)

Set the thrust direction from its unit vector.

pub fn thrust_angles_deg( &self, frame: LocalFrame, ) -> PhysicsResult<Option<(f64, f64)>>

Return the thrust angles in degrees for the provided frame. NOTE: the in-plane and out-of-plane angles differ between the VNC and the RCN frames!

Source

impl Spacecraft

pub fn rss(&self, other: &Self) -> PhysicsResult<(f64, f64, f64)>

Returns the root sum square error between this spacecraft and the other, in kilometers for the position, kilometers per second in velocity, and kilograms in prop