Struct TimeSeries
pub struct TimeSeries { /* private fields */ }Expand description
An iterator of a sequence of evenly spaced Epochs.
(Python documentation hints) :type start: Epoch :type end: Epoch :type step: Duration :type inclusive: bool
Implementations
impl TimeSeries
impl TimeSeries
pub fn exclusive(start: Epoch, end: Epoch, step: Duration) -> TimeSeries ⓘ
pub fn exclusive(start: Epoch, end: Epoch, step: Duration) -> TimeSeries ⓘ
Return an iterator of evenly spaced Epochs, inclusive on start and exclusive on end.
use hifitime::{Epoch, Unit, TimeSeries};
let start = Epoch::from_gregorian_utc_at_midnight(2017, 1, 14);
let end = Epoch::from_gregorian_utc_at_noon(2017, 1, 14);
let step = Unit::Hour * 2;
let time_series = TimeSeries::exclusive(start, end, step);
let mut cnt = 0;
for epoch in time_series {
println!("{}", epoch);
cnt += 1
}
assert_eq!(cnt, 6)pub fn first_epoch(&self) -> Epoch
pub fn first_epoch(&self) -> Epoch
Returns first Epoch of this TimeSeries, without consuming the iterator.
pub fn last_epoch(&self) -> Epoch
pub fn last_epoch(&self) -> Epoch
Returns last Epoch of this TimeSeries, without consuming the iterator.
pub fn inclusive(start: Epoch, end: Epoch, step: Duration) -> TimeSeries ⓘ
pub fn inclusive(start: Epoch, end: Epoch, step: Duration) -> TimeSeries ⓘ
Return an iterator of evenly spaced Epochs, inclusive on start and on end.
use hifitime::{Epoch, Unit, TimeSeries};
let start = Epoch::from_gregorian_utc_at_midnight(2017, 1, 14);
let end = Epoch::from_gregorian_utc_at_noon(2017, 1, 14);
let step = Unit::Hour * 2;
let time_series = TimeSeries::inclusive(start, end, step);
let mut cnt = 0;
for epoch in time_series {
println!("{}", epoch);
cnt += 1
}
assert_eq!(cnt, 7)Examples found in repository?
35fn main() -> Result<(), Box<dyn Error>> {
36 pel::init();
37
38 // ====================== //
39 // === ALMANAC SET UP === //
40 // ====================== //
41
42 // Dynamics models require planetary constants and ephemerides to be defined.
43 // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45 let data_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "examples", "04_lro_od"]
46 .iter()
47 .collect();
48
49 let meta = data_folder.join("lro-dynamics.dhall");
50
51 // Load this ephem in the general Almanac we're using for this analysis.
52 let mut almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
53 .map_err(Box::new)?
54 .process(true)
55 .map_err(Box::new)?;
56
57 let mut moon_pc = almanac.get_planetary_data_from_id(MOON).unwrap();
58 moon_pc.mu_km3_s2 = 4902.74987;
59 almanac.set_planetary_data_from_id(MOON, moon_pc).unwrap();
60
61 let mut earth = almanac.get_planetary_data_from_id(EARTH).unwrap();
62 earth.mu_km3_s2 = 398600.436;
63 almanac.set_planetary_data_from_id(EARTH, earth).unwrap();
64
65 // Save this new kernel for reuse.
66 // In an operational context, this would be part of the "Lock" process, and should not change throughout the mission.
67 almanac
68 .planetary_data
69 .values()
70 .next()
71 .unwrap()
72 .save_as(&data_folder.join("lro-specific.pca"), true)?;
73
74 // Lock the almanac (an Arc is a read only structure).
75 let almanac = Arc::new(almanac);
76
77 // Orbit determination requires a Trajectory structure, which can be saved as parquet file.
78 // In our case, the trajectory comes from the BSP file, so we need to build a Trajectory from the almanac directly.
79 // To query the Almanac, we need to build the LRO frame in the J2000 orientation in our case.
80 // Inspecting the LRO BSP in the ANISE GUI shows us that NASA has assigned ID -85 to LRO.
81 let lro_frame = Frame::from_ephem_j2000(-85);
82
83 // To build the trajectory we need to provide a spacecraft template.
84 let sc_template = Spacecraft::builder()
85 .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0)) // Launch masses
86 .srp(SRPData {
87 // SRP configuration is arbitrary, but we will be estimating it anyway.
88 area_m2: 3.9 * 2.7,
89 coeff_reflectivity: 0.96,
90 })
91 .orbit(Orbit::zero(MOON_J2000)) // Setting a zero orbit here because it's just a template
92 .build();
93 // Now we can build the trajectory from the BSP file.
94 // We'll arbitrarily set the tracking arc to 24 hours with a five second time step.
95 let traj_as_flown = Traj::from_bsp(
96 lro_frame,
97 MOON_J2000,
98 almanac.clone(),
99 sc_template,
100 5.seconds(),
101 Some(Epoch::from_str("2024-01-01 00:00:00 UTC")?),
102 Some(Epoch::from_str("2024-01-02 00:00:00 UTC")?),
103 Aberration::LT,
104 Some("LRO".to_string()),
105 )?;
106
107 println!("{traj_as_flown}");
108
109 // ====================== //
110 // === MODEL MATCHING === //
111 // ====================== //
112
113 // Set up the spacecraft dynamics.
114
115 // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
116 // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
117 let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
118
119 // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
120 // We're using the GRAIL JGGRX model.
121 let mut jggrx_meta = MetaFile {
122 uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
123 crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
124 };
125 // And let's download it if we don't have it yet.
126 jggrx_meta.process(true)?;
127
128 // Build the spherical harmonics.
129 // The harmonics must be computed in the body fixed frame.
130 // We're using the long term prediction of the Moon principal axes frame.
131 let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
132 let sph_harmonics = GravityField::from_stor(
133 almanac.frame_info(moon_pa_frame)?,
134 GravityFieldData::from_shadr(&jggrx_meta.uri, 80, 80, true)?,
135 );
136
137 // Include the spherical harmonics into the orbital dynamics.
138 orbital_dyn.accel_models.push(sph_harmonics);
139
140 // We define the solar radiation pressure, using the default solar flux and accounting only
141 // for the eclipsing caused by the Earth and Moon.
142 // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
143 let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], almanac.clone())?;
144
145 // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
146 // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
147 let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
148
149 println!("{dynamics}");
150
151 // Now we can build the propagator.
152 let setup = Propagator::default_dp78(dynamics.clone());
153
154 // For reference, let's build the trajectory with Nyx's models from that LRO state.
155 let (sim_final, traj_as_sim) = setup
156 .with(*traj_as_flown.first(), almanac.clone())
157 .until_epoch_with_traj(traj_as_flown.last().epoch())?;
158
159 println!("SIM INIT: {:x}", traj_as_flown.first());
160 println!("SIM FINAL: {sim_final:x}");
161 // Compute RIC difference between SIM and LRO ephem
162 let sim_lro_delta = sim_final
163 .orbit
164 .ric_difference(&traj_as_flown.last().orbit)?;
165 println!("{traj_as_sim}");
166 println!(
167 "SIM v LRO - RIC Position (m): {:.3}",
168 sim_lro_delta.radius_km * 1e3
169 );
170 println!(
171 "SIM v LRO - RIC Velocity (m/s): {:.3}",
172 sim_lro_delta.velocity_km_s * 1e3
173 );
174
175 traj_as_sim.ric_diff_to_parquet(
176 &traj_as_flown,
177 "./data/04_output/04_lro_sim_truth_error.parquet",
178 ExportCfg::default(),
179 )?;
180
181 // ==================== //
182 // === OD SIMULATOR === //
183 // ==================== //
184
185 // After quite some time trying to exactly match the model, we still end up with an oscillatory difference on the order of 150 meters between the propagated state
186 // and the truth LRO state.
187
188 // Therefore, we will actually run an estimation from a dispersed LRO state.
189 // The sc_seed is the true LRO state from the BSP.
190 let sc_seed = *traj_as_flown.first();
191
192 // Load the Deep Space Network ground stations.
193 // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
194 let ground_station_file: PathBuf = [
195 env!("CARGO_MANIFEST_DIR"),
196 "examples",
197 "04_lro_od",
198 "dsn-network.yaml",
199 ]
200 .iter()
201 .collect();
202
203 let devices = GroundStation::load_named(ground_station_file)?;
204
205 let mut proc_devices = devices.clone();
206
207 // Increase the noise in the devices to accept more measurements.
208 for gs in proc_devices.values_mut() {
209 if let Some(noise) = &mut gs
210 .stochastic_noises
211 .as_mut()
212 .unwrap()
213 .get_mut(&MeasurementType::Range)
214 {
215 *noise.white_noise.as_mut().unwrap() *= 3.0;
216 }
217 }
218
219 // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
220 // Nyx can build a tracking schedule for you based on the first station with access.
221 let trkconfg_yaml: PathBuf = [
222 env!("CARGO_MANIFEST_DIR"),
223 "examples",
224 "04_lro_od",
225 "tracking-cfg.yaml",
226 ]
227 .iter()
228 .collect();
229
230 let configs: BTreeMap<String, TrkConfig> = TrkConfig::load_named(trkconfg_yaml)?;
231
232 // Build the tracking arc simulation to generate a "standard measurement".
233 let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
234 devices.clone(),
235 traj_as_flown.clone(),
236 configs,
237 123, // Set a seed for reproducibility
238 )?;
239
240 trk.build_schedule(almanac.clone())?;
241 let arc = trk.generate_measurements(almanac.clone())?;
242 // Save the simulated tracking data
243 arc.to_parquet_simple("./data/04_output/04_lro_simulated_tracking.parquet")?;
244
245 // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
246 println!("{arc}");
247
248 // Now that we have simulated measurements, we'll run the orbit determination.
249
250 // ===================== //
251 // === OD ESTIMATION === //
252 // ===================== //
253
254 let sc = SpacecraftUncertainty::builder()
255 .nominal(sc_seed)
256 .frame(LocalFrame::RIC)
257 .x_km(0.5)
258 .y_km(0.5)
259 .z_km(0.5)
260 .vx_km_s(5e-3)
261 .vy_km_s(5e-3)
262 .vz_km_s(5e-3)
263 .build();
264
265 // Build the filter initial estimate, which we will reuse in the filter.
266 let mut initial_estimate = sc.to_estimate()?;
267 initial_estimate.covar *= 3.0;
268
269 println!("== FILTER STATE ==\n{sc_seed:x}\n{initial_estimate}");
270
271 // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
272 let process_noise = ProcessNoise3D::from_velocity_km_s(
273 &[1e-12, 1e-12, 1e-12],
274 1 * Unit::Hour,
275 10 * Unit::Minute,
276 None,
277 );
278
279 println!("{process_noise}");
280
281 // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
282 let odp = SpacecraftKalmanOD::new(
283 setup,
284 KalmanVariant::ReferenceUpdate,
285 Some(ResidRejectCrit::default()),
286 proc_devices,
287 almanac.clone(),
288 )
289 .with_process_noise(process_noise);
290
291 let od_sol = odp.process_arc(initial_estimate, &arc)?;
292
293 let final_est = od_sol.estimates.last().unwrap();
294
295 println!("{final_est}");
296
297 let ric_err = traj_as_flown
298 .at(final_est.epoch())?
299 .orbit
300 .ric_difference(&final_est.orbital_state())?;
301 println!("== RIC at end ==");
302 println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
303 println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
304
305 println!(
306 "Num residuals rejected: #{}",
307 od_sol.rejected_residuals().len()
308 );
309 println!(
310 "Percentage within +/-3: {}",
311 od_sol.residual_ratio_within_threshold(3.0).unwrap()
312 );
313 println!("Ratios normal? {}", od_sol.is_normal(None).unwrap());
314
315 od_sol.to_parquet(
316 "./data/04_output/04_lro_od_results.parquet",
317 ExportCfg::default(),
318 )?;
319
320 // Create the ephemeris
321 let ephem = od_sol.to_ephemeris("LRO rebuilt".to_string());
322 let ephem_start = ephem.start_epoch().unwrap();
323 let ephem_end = ephem.end_epoch().unwrap();
324 // Check that the covariance is PSD throughout the ephemeris by interpolating it.
325 for epoch in TimeSeries::inclusive(ephem_start, ephem_end, Unit::Minute * 5) {
326 ephem
327 .covar_at(
328 epoch,
329 anise::ephemerides::ephemeris::LocalFrame::RIC,
330 &almanac,
331 )
332 .unwrap_or_else(|e| panic!("covar not PSD at {epoch}: {e}"));
333 }
334 // Export as BSP!
335 ephem
336 .write_spice_bsp(-85, "./data/04_output/04_lro_rebuilt.bsp", None)
337 .expect("could not built BSP");
338 let new_almanac = Almanac::default()
339 .load("./data/04_output/04_lro_rebuilt.bsp")
340 .unwrap();
341 new_almanac.describe(None, None, None, None, None, None, None, None);
342 let (spk_start, spk_end) = new_almanac.spk_domain(-85).unwrap();
343
344 assert!((ephem_start - spk_start).abs() < Unit::Microsecond * 1);
345 assert!((ephem_end - spk_end).abs() < Unit::Microsecond * 1);
346
347 // In our case, we have the truth trajectory from NASA.
348 // So we can compute the RIC state difference between the real LRO ephem and what we've just estimated.
349 // Export the OD trajectory first.
350 let od_trajectory = od_sol.to_traj()?;
351 // Build the RIC difference.
352 od_trajectory.ric_diff_to_parquet(
353 &traj_as_flown,
354 "./data/04_output/04_lro_od_truth_error.parquet",
355 ExportCfg::default(),
356 )?;
357
358 Ok(())
359}Trait Implementations
impl Clone for TimeSeries
impl Clone for TimeSeries
fn clone(&self) -> TimeSeries ⓘ
fn clone(&self) -> TimeSeries ⓘ
1.0.0 (const: unstable) · Sourcefn clone_from(&mut self, source: &Self)
fn clone_from(&mut self, source: &Self)
source. Read moreimpl Debug for TimeSeries
impl Debug for TimeSeries
impl Display for TimeSeries
impl Display for TimeSeries
impl DoubleEndedIterator for TimeSeries
impl DoubleEndedIterator for TimeSeries
fn next_back(&mut self) -> Option<Epoch>
fn next_back(&mut self) -> Option<Epoch>
Sourcefn advance_back_by(&mut self, n: usize) -> Result<(), NonZero<usize>>
fn advance_back_by(&mut self, n: usize) -> Result<(), NonZero<usize>>
iter_advance_by)n elements. Read more1.37.0 (const: unstable) · Sourcefn nth_back(&mut self, n: usize) -> Option<Self::Item>
fn nth_back(&mut self, n: usize) -> Option<Self::Item>
nth element from the end of the iterator. Read more1.27.0 (const: unstable) · Sourcefn try_rfold<B, F, R>(&mut self, init: B, f: F) -> R
fn try_rfold<B, F, R>(&mut self, init: B, f: F) -> R
Iterator::try_fold(): it takes
elements starting from the back of the iterator. Read moreimpl ExactSizeIterator for TimeSerieswhere
TimeSeries: Iterator,
impl ExactSizeIterator for TimeSerieswhere
TimeSeries: Iterator,
impl<'py> IntoPyObject<'py> for TimeSeries
impl<'py> IntoPyObject<'py> for TimeSeries
type Target = TimeSeries
type Target = TimeSeries
type Output = Bound<'py, <TimeSeries as IntoPyObject<'py>>::Target>
type Output = Bound<'py, <TimeSeries as IntoPyObject<'py>>::Target>
fn into_pyobject(
self,
py: Python<'py>,
) -> Result<<TimeSeries as IntoPyObject<'py>>::Output, <TimeSeries as IntoPyObject<'py>>::Error>
fn into_pyobject( self, py: Python<'py>, ) -> Result<<TimeSeries as IntoPyObject<'py>>::Output, <TimeSeries as IntoPyObject<'py>>::Error>
impl Iterator for TimeSeries
impl Iterator for TimeSeries
fn size_hint(&self) -> (usize, Option<usize>)
fn size_hint(&self) -> (usize, Option<usize>)
Sourcefn next_chunk<const N: usize>(
&mut self,
) -> Result<[Self::Item; N], IntoIter<Self::Item, N>>where
Self: Sized,
fn next_chunk<const N: usize>(
&mut self,
) -> Result<[Self::Item; N], IntoIter<Self::Item, N>>where
Self: Sized,
iter_next_chunk)N values. Read more1.0.0 (const: unstable) · Sourcefn count(self) -> usizewhere
Self: Sized,
fn count(self) -> usizewhere
Self: Sized,
1.0.0 (const: unstable) · Sourcefn last(self) -> Option<Self::Item>where
Self: Sized,
fn last(self) -> Option<Self::Item>where
Self: Sized,
Sourcefn advance_by(&mut self, n: usize) -> Result<(), NonZero<usize>>
fn advance_by(&mut self, n: usize) -> Result<(), NonZero<usize>>
iter_advance_by)n elements. Read more1.0.0 (const: unstable) · Sourcefn nth(&mut self, n: usize) -> Option<Self::Item>
fn nth(&mut self, n: usize) -> Option<Self::Item>
nth element of the iterator. Read more1.28.0 (const: unstable) · Sourcefn step_by(self, step: usize) -> StepBy<Self>where
Self: Sized,
fn step_by(self, step: usize) -> StepBy<Self>where
Self: Sized,
1.0.0 (const: unstable) · Sourcefn chain<U>(self, other: U) -> Chain<Self, <U as IntoIterator>::IntoIter>
fn chain<U>(self, other: U) -> Chain<Self, <U as IntoIterator>::IntoIter>
1.0.0 (const: unstable) · Sourcefn zip<U>(self, other: U) -> Zip<Self, <U as IntoIterator>::IntoIter>where
Self: Sized,
U: IntoIterator,
fn zip<U>(self, other: U) -> Zip<Self, <U as IntoIterator>::IntoIter>where
Self: Sized,
U: IntoIterator,
Sourcefn intersperse(self, separator: Self::Item) -> Intersperse<Self>
fn intersperse(self, separator: Self::Item) -> Intersperse<Self>
iter_intersperse)separator between items
of the original iterator. Read moreSourcefn intersperse_with<G>(self, separator: G) -> IntersperseWith<Self, G>
fn intersperse_with<G>(self, separator: G) -> IntersperseWith<Self, G>
iter_intersperse)separator
between items of the original iterator. Read more1.0.0 (const: unstable) · Sourcefn map<B, F>(self, f: F) -> Map<Self, F>
fn map<B, F>(self, f: F) -> Map<Self, F>
1.21.0 (const: unstable) · Sourcefn for_each<F>(self, f: F)
fn for_each<F>(self, f: F)
1.0.0 (const: unstable) · Sourcefn filter<P>(self, predicate: P) -> Filter<Self, P>
fn filter<P>(self, predicate: P) -> Filter<Self, P>
1.0.0 (const: unstable) · Sourcefn filter_map<B, F>(self, f: F) -> FilterMap<Self, F>
fn filter_map<B, F>(self, f: F) -> FilterMap<Self, F>
1.0.0 (const: unstable) · Sourcefn enumerate(self) -> Enumerate<Self>where
Self: Sized,
fn enumerate(self) -> Enumerate<Self>where
Self: Sized,
1.0.0 (const: unstable) · Sourcefn skip_while<P>(self, predicate: P) -> SkipWhile<Self, P>
fn skip_while<P>(self, predicate: P) -> SkipWhile<Self, P>
1.0.0 (const: unstable) · Sourcefn take_while<P>(self, predicate: P) -> TakeWhile<Self, P>
fn take_while<P>(self, predicate: P) -> TakeWhile<Self, P>
1.57.0 (const: unstable) · Sourcefn map_while<B, P>(self, predicate: P) -> MapWhile<Self, P>
fn map_while<B, P>(self, predicate: P) -> MapWhile<Self, P>
1.0.0 (const: unstable) · Sourcefn skip(self, n: usize) -> Skip<Self>where
Self: Sized,
fn skip(self, n: usize) -> Skip<Self>where
Self: Sized,
n elements. Read more1.0.0 (const: unstable) · Sourcefn take(self, n: usize) -> Take<Self>where
Self: Sized,
fn take(self, n: usize) -> Take<Self>where
Self: Sized,
n elements, or fewer
if the underlying iterator ends sooner. Read more1.0.0 (const: unstable) · Sourcefn scan<St, B, F>(self, initial_state: St, f: F) -> Scan<Self, St, F>
fn scan<St, B, F>(self, initial_state: St, f: F) -> Scan<Self, St, F>
1.0.0 (const: unstable) · Sourcefn flat_map<U, F>(self, f: F) -> FlatMap<Self, U, F>
fn flat_map<U, F>(self, f: F) -> FlatMap<Self, U, F>
1.29.0 (const: unstable) · Sourcefn flatten(self) -> Flatten<Self>
fn flatten(self) -> Flatten<Self>
Sourcefn map_windows<F, R, const N: usize>(self, f: F) -> MapWindows<Self, F, N>
fn map_windows<F, R, const N: usize>(self, f: F) -> MapWindows<Self, F, N>
iter_map_windows)f for each contiguous window of size N over
self and returns an iterator over the outputs of f. Like slice::windows(),
the windows during mapping overlap as well. Read more1.0.0 (const: unstable) · Sourcefn inspect<F>(self, f: F) -> Inspect<Self, F>
fn inspect<F>(self, f: F) -> Inspect<Self, F>
1.0.0 (const: unstable) · Sourcefn by_ref(&mut self) -> &mut Selfwhere
Self: Sized,
fn by_ref(&mut self) -> &mut Selfwhere
Self: Sized,
Iterator. Read more1.0.0 (const: unstable) · Sourcefn collect<B>(self) -> B
fn collect<B>(self) -> B
Sourcefn try_collect<B>(
&mut self,
) -> <<Self::Item as Try>::Residual as Residual<B>>::TryType
fn try_collect<B>( &mut self, ) -> <<Self::Item as Try>::Residual as Residual<B>>::TryType
iterator_try_collect)Sourcefn collect_into<E>(self, collection: &mut E) -> &mut E
fn collect_into<E>(self, collection: &mut E) -> &mut E
iter_collect_into)1.0.0 (const: unstable) · Sourcefn partition<B, F>(self, f: F) -> (B, B)
fn partition<B, F>(self, f: F) -> (B, B)
Sourcefn partition_in_place<'a, T, P>(self, predicate: P) -> usize
fn partition_in_place<'a, T, P>(self, predicate: P) -> usize
iter_partition_in_place)true precede all those that return false.
Returns the number of true elements found. Read moreSourcefn is_partitioned<P>(self, predicate: P) -> bool
fn is_partitioned<P>(self, predicate: P) -> bool
iter_is_partitioned)true precede all those that return false. Read more1.27.0 (const: unstable) · Sourcefn try_fold<B, F, R>(&mut self, init: B, f: F) -> R
fn try_fold<B, F, R>(&mut self, init: B, f: F) -> R
1.27.0 (const: unstable) · Sourcefn try_for_each<F, R>(&mut self, f: F) -> R
fn try_for_each<F, R>(&mut self, f: F) -> R
1.0.0 (const: unstable) · Sourcefn fold<B, F>(self, init: B, f: F) -> B
fn fold<B, F>(self, init: B, f: F) -> B
1.51.0 (const: unstable) · Sourcefn reduce<F>(self, f: F) -> Option<Self::Item>
fn reduce<F>(self, f: F) -> Option<Self::Item>
Sourcefn try_reduce<R>(
&mut self,
f: impl FnMut(Self::Item, Self::Item) -> R,
) -> <<R as Try>::Residual as Residual<Option<<R as Try>::Output>>>::TryType
fn try_reduce<R>( &mut self, f: impl FnMut(Self::Item, Self::Item) -> R, ) -> <<R as Try>::Residual as Residual<Option<<R as Try>::Output>>>::TryType
iterator_try_reduce)1.0.0 (const: unstable) · Sourcefn all<F>(&mut self, f: F) -> bool
fn all<F>(&mut self, f: F) -> bool
1.0.0 (const: unstable) · Sourcefn any<F>(&mut self, f: F) -> bool
fn any<F>(&mut self, f: F) -> bool
1.0.0 (const: unstable) · Sourcefn find<P>(&mut self, predicate: P) -> Option<Self::Item>
fn find<P>(&mut self, predicate: P) -> Option<Self::Item>
1.30.0 (const: unstable) · Sourcefn find_map<B, F>(&mut self, f: F) -> Option<B>
fn find_map<B, F>(&mut self, f: F) -> Option<B>
Sourcefn try_find<R>(
&mut self,
f: impl FnMut(&Self::Item) -> R,
) -> <<R as Try>::Residual as Residual<Option<Self::Item>>>::TryType
fn try_find<R>( &mut self, f: impl FnMut(&Self::Item) -> R, ) -> <<R as Try>::Residual as Residual<Option<Self::Item>>>::TryType
try_find)1.0.0 (const: unstable) · Sourcefn position<P>(&mut self, predicate: P) -> Option<usize>
fn position<P>(&mut self, predicate: P) -> Option<usize>
1.0.0 (const: unstable) · Sourcefn rposition<P>(&mut self, predicate: P) -> Option<usize>
fn rposition<P>(&mut self, predicate: P) -> Option<usize>
1.0.0 (const: unstable) · Sourcefn max(self) -> Option<Self::Item>
fn max(self) -> Option<Self::Item>
1.0.0 (const: unstable) · Sourcefn min(self) -> Option<Self::Item>
fn min(self) -> Option<Self::Item>
1.6.0 (const: unstable) · Sourcefn max_by_key<B, F>(self, f: F) -> Option<Self::Item>
fn max_by_key<B, F>(self, f: F) -> Option<Self::Item>
1.15.0 (const: unstable) · Sourcefn max_by<F>(self, compare: F) -> Option<Self::Item>
fn max_by<F>(self, compare: F) -> Option<Self::Item>
1.6.0 (const: unstable) · Sourcefn min_by_key<B, F>(self, f: F) -> Option<Self::Item>
fn min_by_key<B, F>(self, f: F) -> Option<Self::Item>
1.15.0 (const: unstable) · Sourcefn min_by<F>(self, compare: F) -> Option<Self::Item>
fn min_by<F>(self, compare: F) -> Option<Self::Item>
1.0.0 (const: unstable) · Sourcefn rev(self) -> Rev<Self>where
Self: Sized + DoubleEndedIterator,
fn rev(self) -> Rev<Self>where
Self: Sized + DoubleEndedIterator,
1.0.0 (const: unstable) · Sourcefn unzip<A, B, FromA, FromB>(self) -> (FromA, FromB)
fn unzip<A, B, FromA, FromB>(self) -> (FromA, FromB)
1.36.0 (const: unstable) · Sourcefn copied<'a, T>(self) -> Copied<Self>
fn copied<'a, T>(self) -> Copied<Self>
1.0.0 (const: unstable) · Sourcefn cycle(self) -> Cycle<Self>
fn cycle(self) -> Cycle<Self>
Sourcefn array_chunks<const N: usize>(self) -> ArrayChunks<Self, N>where
Self: Sized,
fn array_chunks<const N: usize>(self) -> ArrayChunks<Self, N>where
Self: Sized,
iter_array_chunks)N elements of the iterator at a time. Read more1.11.0 (const: unstable) · Sourcefn product<P>(self) -> P
fn product<P>(self) -> P
Sourcefn cmp_by<I, F>(self, other: I, cmp: F) -> Ordering
fn cmp_by<I, F>(self, other: I, cmp: F) -> Ordering
iter_order_by)Iterator with those
of another with respect to the specified comparison function. Read more1.5.0 (const: unstable) · Sourcefn partial_cmp<I>(self, other: I) -> Option<Ordering>
fn partial_cmp<I>(self, other: I) -> Option<Ordering>
PartialOrd elements of
this Iterator with those of another. The comparison works like short-circuit
evaluation, returning a result without comparing the remaining elements.
As soon as an order can be determined, the evaluation stops and a result is returned. Read moreSourcefn partial_cmp_by<I, F>(self, other: I, partial_cmp: F) -> Option<Ordering>where
Self: Sized,
I: IntoIterator,
F: FnMut(Self::Item, <I as IntoIterator>::Item) -> Option<Ordering>,
fn partial_cmp_by<I, F>(self, other: I, partial_cmp: F) -> Option<Ordering>where
Self: Sized,
I: IntoIterator,
F: FnMut(Self::Item, <I as IntoIterator>::Item) -> Option<Ordering>,
iter_order_by)Iterator with those
of another with respect to the specified comparison function. Read moreSourcefn eq_by<I, F>(self, other: I, eq: F) -> bool
fn eq_by<I, F>(self, other: I, eq: F) -> bool
iter_order_by)1.5.0 (const: unstable) · Sourcefn lt<I>(self, other: I) -> bool
fn lt<I>(self, other: I) -> bool
Iterator are lexicographically
less than those of another. Read more1.5.0 (const: unstable) · Sourcefn le<I>(self, other: I) -> bool
fn le<I>(self, other: I) -> bool
Iterator are lexicographically
less or equal to those of another. Read more1.5.0 (const: unstable) · Sourcefn gt<I>(self, other: I) -> bool
fn gt<I>(self, other: I) -> bool
Iterator are lexicographically
greater than those of another. Read more1.5.0 (const: unstable) · Sourcefn ge<I>(self, other: I) -> bool
fn ge<I>(self, other: I) -> bool
Iterator are lexicographically
greater than or equal to those of another. Read more1.82.0 (const: unstable) · Sourcefn is_sorted(self) -> bool
fn is_sorted(self) -> bool
1.82.0 (const: unstable) · Sourcefn is_sorted_by<F>(self, compare: F) -> bool
fn is_sorted_by<F>(self, compare: F) -> bool
impl LowerExp for TimeSeries
impl LowerExp for TimeSeries
impl LowerHex for TimeSeries
impl LowerHex for TimeSeries
impl Octal for TimeSeries
impl Octal for TimeSeries
impl PartialEq for TimeSeries
impl PartialEq for TimeSeries
impl Pointer for TimeSeries
impl Pointer for TimeSeries
impl PyClass for TimeSeries
impl PyClass for TimeSeries
impl PyTypeInfo for TimeSeries
impl PyTypeInfo for TimeSeries
const NAME: &'static str = <Self as ::pyo3::PyClass>::NAME
const NAME: &'static str = <Self as ::pyo3::PyClass>::NAME
prefer using ::type_object(py).name() to get the correct runtime value
const MODULE: Option<&'static str> = <Self as ::pyo3::impl_::pyclass::PyClassImpl>::MODULE
const MODULE: Option<&'static str> = <Self as ::pyo3::impl_::pyclass::PyClassImpl>::MODULE
prefer using ::type_object(py).module() to get the correct runtime value
fn type_object_raw(py: Python<'_>) -> *mut PyTypeObject
fn type_object_raw(py: Python<'_>) -> *mut PyTypeObject
fn type_object(py: Python<'_>) -> Bound<'_, PyType>
fn type_object(py: Python<'_>) -> Bound<'_, PyType>
fn is_type_of(object: &Bound<'_, PyAny>) -> bool
fn is_type_of(object: &Bound<'_, PyAny>) -> bool
object is an instance of this type or a subclass of this type.fn is_exact_type_of(object: &Bound<'_, PyAny>) -> bool
fn is_exact_type_of(object: &Bound<'_, PyAny>) -> bool
object is an instance of this type.impl UpperExp for TimeSeries
impl UpperExp for TimeSeries
impl UpperHex for TimeSeries
impl UpperHex for TimeSeries
impl DerefToPyAny for TimeSeries
impl Eq for TimeSeries
impl StructuralPartialEq for TimeSeries